from __future__ import annotations

import logging
from typing import Dict, Iterator, List, Optional, Sequence, Union, cast

from opentrons_shared_data.errors.exceptions import CommandPreconditionViolated
from opentrons_shared_data.labware.types import LabwareDefinition
from opentrons_shared_data.module.types import ModuleModel, ModuleType

from .core.common import (
    AbsorbanceReaderCore,
    FlexStackerCore,
    HeaterShakerCore,
    LabwareCore,
    MagneticBlockCore,
    MagneticModuleCore,
    ModuleCore,
    ProtocolCore,
    TemperatureModuleCore,
    ThermocyclerCore,
    VacuumModuleCore,
)
from .core.core_map import LoadedCoreMap
from .core.engine import ENGINE_CORE_API_VERSION
from .core.legacy.legacy_labware_core import LegacyLabwareCore as LegacyLabwareCore
from .core.legacy.legacy_module_core import LegacyModuleCore
from .core.legacy.module_geometry import ModuleGeometry as LegacyModuleGeometry
from .labware import Labware
from .module_validation_and_errors import (
    validate_heater_shaker_speed,
    validate_heater_shaker_temperature,
)
from opentrons.drivers.thermocycler.driver import BLOCK_VOL_MAX, BLOCK_VOL_MIN
from opentrons.legacy_broker import LegacyBroker
from opentrons.legacy_commands import module_commands as cmds
from opentrons.legacy_commands.publisher import CommandPublisher, publish
from opentrons.protocol_engine.types import ABSMeasureMode
from opentrons.protocols.api_support.types import APIVersion, ThermocyclerStep
from opentrons.protocols.api_support.util import (
    APIVersionError,
    UnsupportedAPIError,
    requires_version,
)

from . import validation  # isort: skip  # Imported after other protocol_api imports to avoid circular import
from .tasks import Task  # isort: skip

_MAGNETIC_MODULE_HEIGHT_PARAM_REMOVED_IN = APIVersion(2, 14)


_log = logging.getLogger(__name__)


class ModuleContext(CommandPublisher):
    """A connected module in the protocol.

    *New in version 2.0*
    """

    def __init__(
        self,
        core: ModuleCore,
        protocol_core: ProtocolCore,
        core_map: LoadedCoreMap,
        api_version: APIVersion,
        broker: LegacyBroker,
    ) -> None:
        super().__init__(broker=broker)
        self._core = core
        self._protocol_core = protocol_core
        self._core_map = core_map
        self._api_version = api_version

    @property
    @requires_version(2, 0)
    def api_version(self) -> APIVersion:
        return self._api_version

    @property
    @requires_version(2, 14)
    def model(self) -> ModuleModel:
        """Get the module's model identifier."""
        return cast(ModuleModel, self._core.get_model().value)

    @property
    @requires_version(2, 14)
    def type(self) -> ModuleType:
        """Get the module's general type identifier."""
        return self._get_type()

    def _get_type(self) -> ModuleType:
        return self._core.MODULE_TYPE.value

    @requires_version(2, 0)
    def load_labware_object(self, labware: Labware) -> Labware:
        """
        Specify the presence of a piece of labware on the module.

        Args:
            labware: The labware object. This object should be already
                initialized and its parent should be set to this module's
                geometry. To initialize and load a labware onto the module in
                one step, see `load_labware()`.

        Returns:
            The properly-linked labware object.

        *Deprecated in version 2.14:* Use `load_labware()` or `load_labware_by_definition()`.
        """
        if not isinstance(self._core, LegacyModuleCore):
            raise UnsupportedAPIError(
                api_element="`ModuleContext.load_labware_object`",
                since_version="2.14",
                extra_message="Use `ModuleContext.load_labware` or `load_labware_by_definition` instead.",
            )

        _log.warning(
            "`ModuleContext.load_labware_object` is an internal, deprecated method. Use `ModuleContext.load_labware` or `load_labware_by_definition` instead."
        )

        assert labware.parent == self._core.geometry, (
            "Labware is not configured with this module as its parent"
        )

        return self._core.geometry.add_labware(labware)

    def load_labware(  # noqa: C901
        self,
        name: str,
        label: Optional[str] = None,
        namespace: Optional[str] = None,
        version: Optional[int] = None,
        adapter: Optional[str] = None,
        lid: Optional[str] = None,
        *,
        adapter_namespace: Optional[str] = None,
        adapter_version: Optional[int] = None,
        lid_namespace: Optional[str] = None,
        lid_version: Optional[int] = None,
    ) -> Labware:
        """
        Load a labware onto the module using its load parameters.

        The parameters of this function behave like those of
        [`load_labware()`][opentrons.protocol_api.ProtocolContext.load_labware]
        (which loads labware directly onto the deck). Note that the parameter
        `name` here corresponds to `load_name` on the `ProtocolContext`
        function.

        Returns:
            The initialized and loaded labware object.

        *New in version 2.1:* The `label`, `namespace`, and `version` parameters.

        *New in version 2.26:* The `adapter_namespace`, `adapter_version`,
        `lid_namespace`, and `lid_version` parameters.
        """
        if self._api_version < APIVersion(2, 1) and (
            label is not None or namespace is not None or version != 1
        ):
            _log.warning(
                f"You have specified API {self.api_version}, but you "
                "are trying to utilize new load_labware parameters in 2.1"
            )

        if self._api_version < validation.NAMESPACE_VERSION_ADAPTER_LID_VERSION_GATE:
            if adapter_namespace is not None:
                raise APIVersionError(
                    api_element="The `adapter_namespace` parameter",
                    until_version=str(
                        validation.NAMESPACE_VERSION_ADAPTER_LID_VERSION_GATE
                    ),
                    current_version=str(self._api_version),
                )
            if adapter_version is not None:
                raise APIVersionError(
                    api_element="The `adapter_version` parameter",
                    until_version=str(
                        validation.NAMESPACE_VERSION_ADAPTER_LID_VERSION_GATE
                    ),
                    current_version=str(self._api_version),
                )
            if lid_namespace is not None:
                raise APIVersionError(
                    api_element="The `lid_namespace` parameter",
                    until_version=str(
                        validation.NAMESPACE_VERSION_ADAPTER_LID_VERSION_GATE
                    ),
                    current_version=str(self._api_version),
                )
            if lid_version is not None:
                raise APIVersionError(
                    api_element="The `lid_version` parameter",
                    until_version=str(
                        validation.NAMESPACE_VERSION_ADAPTER_LID_VERSION_GATE
                    ),
                    current_version=str(self._api_version),
                )

        load_location: Union[ModuleCore, LabwareCore]
        if adapter is not None:
            if self._api_version < APIVersion(2, 15):
                raise APIVersionError(
                    api_element="Loading a labware on an adapter",
                    until_version="2.15",
                    current_version=f"{self._api_version}",
                )

            if (
                self._api_version
                < validation.NAMESPACE_VERSION_ADAPTER_LID_VERSION_GATE
            ):
                checked_adapter_namespace = namespace
                checked_adapter_version = None
            else:
                checked_adapter_namespace = adapter_namespace
                checked_adapter_version = adapter_version

            loaded_adapter = self.load_adapter(
                name=adapter,
                namespace=checked_adapter_namespace,
                version=checked_adapter_version,
            )
            load_location = loaded_adapter._core
        else:
            load_location = self._core

        name = validation.ensure_lowercase_name(name)

        # todo(mm, 2024-11-08): This check belongs in opentrons.protocol_api.core.engine.deck_conflict.
        # We're currently doing it here, at the ModuleContext level, for consistency with what
        # ProtocolContext.load_labware() does. (It should also be moved to the deck_conflict module.)
        if self._get_type() == "absorbanceReaderType":
            if cast(AbsorbanceReaderCore, self._core).is_lid_on():
                raise CommandPreconditionViolated(
                    f"Cannot load {name} onto the Absorbance Reader Module when its lid is closed."
                )

        labware_core = self._protocol_core.load_labware(
            load_name=name,
            label=label if label is None else str(label),
            namespace=namespace,
            version=version,
            location=load_location,
        )
        if lid is not None:
            if self._api_version < validation.LID_STACK_VERSION_GATE:
                raise APIVersionError(
                    api_element="Loading a lid on a Labware",
                    until_version="2.23",
                    current_version=f"{self._api_version}",
                )

            if (
                self._api_version
                < validation.NAMESPACE_VERSION_ADAPTER_LID_VERSION_GATE
            ):
                checked_lid_namespace = namespace
                checked_lid_version = None
            else:
                checked_lid_namespace = lid_namespace
                checked_lid_version = lid_version

            self._protocol_core.load_lid(
                load_name=lid,
                location=labware_core,
                namespace=checked_lid_namespace,
                version=checked_lid_version,
            )

        if isinstance(self._core, LegacyModuleCore) and isinstance(
            labware_core, LegacyLabwareCore
        ):
            labware = self._core.add_labware_core(labware_core)
        else:
            labware = Labware(
                core=labware_core,
                api_version=self._api_version,
                protocol_core=self._protocol_core,
                core_map=self._core_map,
            )

        self._core_map.add(labware_core, labware)

        return labware

    @requires_version(2, 0)
    def load_labware_from_definition(
        self, definition: LabwareDefinition, label: Optional[str] = None
    ) -> Labware:
        """Load a labware onto the module using an inline definition.

        Args:
            definition: The labware definition.
            label (str): An optional special name to give the labware. If
                specified, this is the name the labware will appear as in the
                run log and the calibration view in the Opentrons App.

        Returns:
            The initialized and loaded labware object.
        """
        load_params = self._protocol_core.add_labware_definition(definition)

        return self.load_labware(
            name=load_params.load_name,
            namespace=load_params.namespace,
            version=load_params.version,
            label=label,
        )

    @requires_version(2, 1)
    def load_labware_by_name(
        self,
        name: str,
        label: Optional[str] = None,
        namespace: Optional[str] = None,
        version: Optional[int] = None,
    ) -> Labware:
        """
        *Deprecated in version 2.0:*
            Use `load_labware` instead.
        """
        _log.warning("load_labware_by_name is deprecated. Use load_labware instead.")
        return self.load_labware(
            name=name,
            label=label,
            namespace=namespace,
            version=version,
        )

    @requires_version(2, 15)
    def load_adapter(
        self,
        name: str,
        namespace: Optional[str] = None,
        version: Optional[int] = None,
    ) -> Labware:
        """Load an adapter onto the module using its load parameters.

        The parameters of this function behave like those of
        [`load_adapter()`][opentrons.protocol_api.ProtocolContext.load_adapter]
        (which loads adapters directly onto the deck). Note that the parameter
        `name` here corresponds to `load_name` on the `ProtocolContext`
        function.

        Returns:
            The initialized and loaded adapter object.
        """
        labware_core = self._protocol_core.load_adapter(
            load_name=name,
            namespace=namespace,
            version=version,
            location=self._core,
        )

        if isinstance(self._core, LegacyModuleCore) and isinstance(
            labware_core, LegacyLabwareCore
        ):
            adapter = self._core.add_labware_core(labware_core)
        else:
            adapter = Labware(
                core=labware_core,
                api_version=self._api_version,
                protocol_core=self._protocol_core,
                core_map=self._core_map,
            )

        self._core_map.add(labware_core, adapter)

        return adapter

    @requires_version(2, 15)
    def load_adapter_from_definition(self, definition: LabwareDefinition) -> Labware:
        """
        Load an adapter onto the module using an inline definition.

        Args:
            definition: The labware definition.

        Returns:
            The initialized and loaded labware object.
        """
        load_params = self._protocol_core.add_labware_definition(definition)

        return self.load_adapter(
            name=load_params.load_name,
            namespace=load_params.namespace,
            version=load_params.version,
        )

    @property
    @requires_version(2, 0)
    def labware(self) -> Optional[Labware]:
        """The labware (if any) present on this module."""
        labware_core = self._protocol_core.get_labware_on_module(self._core)
        return self._core_map.get(labware_core)

    @property
    @requires_version(2, 14)
    def parent(self) -> str:
        """The name of the slot the module is on.

        On a Flex, this will be like `"D1"`. On an OT-2, this will be like `"1"`.
        See [Deck Slots](../deck-slots.md).
        """
        return self._core.get_deck_slot_id()

    @property
    @requires_version(2, 0)
    def geometry(self) -> LegacyModuleGeometry:
        """The object representing the module as an item on the deck.

        *Deprecated in version 2.14:*
            Use properties of the `ModuleContext` instead,
            like `model` and `type`.
        """
        if isinstance(self._core, LegacyModuleCore):
            return self._core.geometry

        raise UnsupportedAPIError(
            api_element="`ModuleContext.geometry`",
            since_version="2.14",
            extra_message="Use properties of the `ModuleContext` itself.",
        )

    def __repr__(self) -> str:
        class_name = self.__class__.__name__
        display_name = self._core.get_display_name()
        location = self._core.get_deck_slot().id

        return f"{class_name} at {display_name} on {location} lw {self.labware}"


class TemperatureModuleContext(ModuleContext):
    """
    An object representing a connected Temperature Module.

    It should not be instantiated directly; instead, it should be
    created through [`ProtocolContext.load_module()`][opentrons.protocol_api.ProtocolContext.load_module].

    *New in version 2.0*
    """

    _core: TemperatureModuleCore

    @property
    @requires_version(2, 14)
    def serial_number(self) -> str:
        """Get the module's unique hardware serial number."""
        return self._core.get_serial_number()

    @publish(command=cmds.tempdeck_set_temp)
    @requires_version(2, 0)
    def set_temperature(self, celsius: float) -> None:
        """Set a target temperature and wait until the module reaches the target.

        No other protocol commands will execute while waiting for the temperature.

        Args:
            celsius: A value between 4 and 95, representing the target temperature in °C.
        """
        self._core.set_target_temperature(celsius)
        self._core.wait_for_target_temperature()

    @publish(command=cmds.tempdeck_set_temp)
    @requires_version(2, 3)
    def start_set_temperature(self, celsius: float) -> Task:
        """Sets the Temperature Module's target temperature and returns immediately without waiting for the module to reach the target. Allows the protocol to proceed while the Temperature Module heats.

        *Changed in version 2.27:* Returns a task object that represents
        concurrent heating. Pass the task object to
        [`wait_for_tasks()`][opentrons.protocol_api.ProtocolContext.wait_for_tasks]
        to wait for the module to finish heating.

        In API version 2.26 or below, this function returns `None`.

        Args:
            celsius: A value between 4 and 95, representing the target
                temperature in °C.
        """
        task = self._core.set_target_temperature(celsius)
        if self._api_version >= APIVersion(2, 27):
            return Task(api_version=self._api_version, core=task)
        else:
            return cast(Task, None)

    @publish(command=cmds.tempdeck_await_temp)
    @requires_version(2, 3)
    def await_temperature(self, celsius: float) -> None:
        """Wait until module reaches temperature.

        Args:
            celsius: A value between 4 and 95, representing the target temperature
                in °C.
        """
        self._core.wait_for_target_temperature(celsius)

    @publish(command=cmds.tempdeck_deactivate)
    @requires_version(2, 0)
    def deactivate(self) -> None:
        """Stop heating or cooling, and turn off the fan."""
        self._core.deactivate()

    @property
    @requires_version(2, 0)
    def temperature(self) -> float:
        """The current temperature of the Temperature Module's deck in °C.

        Returns `0` in simulation if no target temperature has been set.
        """
        return self._core.get_current_temperature()

    @property
    @requires_version(2, 0)
    def target(self) -> Optional[float]:
        """The target temperature of the Temperature Module's deck in °C.

        Returns `None` if no target has been set.
        """
        return self._core.get_target_temperature()

    @property
    @requires_version(2, 3)
    def status(self) -> str:
        """One of four possible temperature statuses:

        - `holding at target`: The module has reached its target temperature
          and is actively maintaining that temperature.
        - `cooling`: The module is cooling to a target temperature.
        - `heating`: The module is heating to a target temperature.
        - `idle`: The module has been deactivated.
        """
        return self._core.get_status().value


class MagneticModuleContext(ModuleContext):
    """
    An object representing a connected Magnetic Module.

    It should not be instantiated directly; instead, it should be
    created through [`ProtocolContext.load_module()`][opentrons.protocol_api.ProtocolContext.load_module].

    *New in version 2.0*
    """

    _core: MagneticModuleCore

    @property
    @requires_version(2, 14)
    def serial_number(self) -> str:
        """Get the module's unique hardware serial number."""
        return self._core.get_serial_number()

    @publish(command=cmds.magdeck_calibrate)
    @requires_version(2, 0)
    def calibrate(self) -> None:
        """Calibrate the Magnetic Module.

        *Deprecated in version 2.14:* This method is unnecessary; remove any usage.
        """
        if self._api_version < ENGINE_CORE_API_VERSION:
            _log.warning(
                "`MagneticModuleContext.calibrate` doesn't do anything useful"
                " and will be removed in Protocol API version 2.14 and higher."
            )
            self._core._sync_module_hardware.calibrate()  # type: ignore[attr-defined]
        else:
            raise UnsupportedAPIError(
                api_element="`MagneticModuleContext.calibrate`",
                since_version="2.14",
            )

    @publish(command=cmds.magdeck_engage)
    @requires_version(2, 0)
    def engage(
        self,
        height: Optional[float] = None,
        offset: Optional[float] = None,
        height_from_base: Optional[float] = None,
    ) -> None:
        """
        Raise the Magnetic Module's magnets. You can specify how high the magnets
        should move:

        - No parameter: Move to the default height for the loaded labware. If
          the loaded labware has no default, or if no labware is loaded, this will
          raise an error.

        - `height_from_base`: Move this many millimeters above the bottom
          of the labware. Acceptable values are between `0` and `25`.

            This is the recommended way to adjust the magnets' height.

            *New in version 2.2*

        - `offset`: Move this many millimeters above (positive value) or below
          (negative value) the default height for the loaded labware. The sum of
          the default height and `offset` must be between 0 and 25.

        - `height`: Intended to move this many millimeters above the magnets'
          home position. However, depending on the generation of module and the loaded
          labware, this may produce unpredictable results. You should normally use
          `height_from_base` instead.

            *Removed in version 2.14*

        You shouldn't specify more than one of these parameters. However, if you do,
        their order of precedence is `height`, then `height_from_base`, then `offset`.
        """
        if height is not None:
            if self._api_version >= _MAGNETIC_MODULE_HEIGHT_PARAM_REMOVED_IN:
                raise UnsupportedAPIError(
                    api_element="The height parameter of MagneticModuleContext.engage()",
                    since_version=f"{_MAGNETIC_MODULE_HEIGHT_PARAM_REMOVED_IN}",
                    current_version=f"{self._api_version}",
                    extra_message="Use offset or height_from_base.",
                )
            self._core.engage(height_from_home=height)

        # This version check has a bug:
        # if the caller sets height_from_base in an API version that's too low,
        # we will silently ignore it instead of raising APIVersionError.
        # Leaving this unfixed because we haven't thought through
        # how to do backwards-compatible fixes to our version checking itself.
        elif height_from_base is not None and self._api_version >= APIVersion(2, 2):
            self._core.engage(height_from_base=height_from_base)

        else:
            self._core.engage_to_labware(
                offset=offset or 0,
                preserve_half_mm=self._api_version < APIVersion(2, 3),
            )

    @publish(command=cmds.magdeck_disengage)
    @requires_version(2, 0)
    def disengage(self) -> None:
        """Lower the magnets back into the Magnetic Module."""
        self._core.disengage()

    @property
    @requires_version(2, 0)
    def status(self) -> str:
        """The status of the module, either `engaged` or `disengaged`."""
        return self._core.get_status().value


class ThermocyclerContext(ModuleContext):
    """An object representing a connected Thermocycler Module.

    It should not be instantiated directly; instead, it should be
    created through [`ProtocolContext.load_module()`][opentrons.protocol_api.ProtocolContext.load_module].

    *New in version 2.0*
    """

    _core: ThermocyclerCore

    @property
    @requires_version(2, 14)
    def serial_number(self) -> str:
        """Get the module's unique hardware serial number."""
        return self._core.get_serial_number()

    @publish(command=cmds.thermocycler_open)
    @requires_version(2, 0)
    def open_lid(self) -> str:
        """Open the lid."""
        return self._core.open_lid().value

    @publish(command=cmds.thermocycler_close)
    @requires_version(2, 0)
    def close_lid(self) -> str:
        """Close the lid."""
        return self._core.close_lid().value

    @publish(command=cmds.thermocycler_set_block_temp)
    @requires_version(2, 0)
    def set_block_temperature(
        self,
        temperature: float,
        hold_time_seconds: Optional[float] = None,
        hold_time_minutes: Optional[float] = None,
        ramp_rate: Optional[float] = None,
        block_max_volume: Optional[float] = None,
    ) -> None:
        """
        Set the target temperature for the well block, in °C.

        Args:
            temperature: A value between 4 and 99, representing the target
                temperature in °C.
            hold_time_minutes: The number of minutes to hold, after reaching
                `temperature`, before proceeding to the next command. If
                `hold_time_seconds` is also specified, the times are added
                together.
            hold_time_seconds: The number of seconds to hold, after reaching
                `temperature`, before proceeding to the next command. If
                `hold_time_minutes` is also specified, the times are added
                together.
            block_max_volume: The greatest volume of liquid contained in any
                individual well of the loaded labware, in µL. If not specified,
                the default is 25 µL.

                *Changed in version 2.27:* After API
                version 2.27 it will attempt to use the liquid tracking of the
                labware first and then fall back to the 25 if there is no
                probed or loaded liquid.

        !!! note
            If `hold_time_minutes` and `hold_time_seconds` are not specified,
            the Thermocycler will proceed to the next command immediately after
            `temperature` is reached.
        """
        seconds = validation.ensure_hold_time_seconds(
            seconds=hold_time_seconds, minutes=hold_time_minutes
        )
        if self._api_version >= APIVersion(2, 27) and block_max_volume is None:
            block_max_volume = self._get_current_labware_max_vol()
        if self._api_version < APIVersion(2, 28) and ramp_rate:
            # because the argument was always there but didn't work, continue to swallow this arg on
            # old api versions.
            ramp_rate = None
        self._core.set_target_block_temperature(
            celsius=temperature,
            hold_time_seconds=seconds,
            block_max_volume=block_max_volume,
            ramp_rate=ramp_rate,
        )
        self._core.wait_for_block_temperature()

    @publish(command=cmds.thermocycler_start_set_block_temp)
    @requires_version(2, 27)
    def start_set_block_temperature(
        self,
        temperature: float,
        ramp_rate: Optional[float] = None,
        block_max_volume: Optional[float] = None,
    ) -> Task:
        """Sets the target temperature for the Thermocycler Module's well block, in °C.

        Returns a [`Task`][opentrons.protocol_api.Task] object that represents concurrent heating.
        Pass the task object to [`ProtocolContext.wait_for_tasks()`][opentrons.protocol_api.ProtocolContext.wait_for_tasks]
        to wait for the preheat to complete.

        Args:
            temperature: A value between 4 and 99, representing the target
                temperature in °C.
            block_max_volume: The greatest volume of liquid contained in any
                individual well of the loaded labware, in µL. If not specified,
                the default is 25 µL.

        *Changed in version 2.27:* In API version
        2.27 and newer, the API will first attempt to use the liquid tracking in labware, then default to 25 µL if the protocol lacks probed or loaded
        liquid information.
        """

        if block_max_volume is None:
            block_max_volume = self._get_current_labware_max_vol()
        if self._api_version < APIVersion(2, 28) and ramp_rate:
            # because the argument was always there but didn't work, continue to swallow this arg on
            # old api versions.
            ramp_rate = None
        task = self._core.start_set_target_block_temperature(
            celsius=temperature,
            block_max_volume=block_max_volume,
            ramp_rate=ramp_rate,
        )
        return Task(api_version=self._api_version, core=task)

    @publish(command=cmds.thermocycler_set_lid_temperature)
    @requires_version(2, 0)
    def set_lid_temperature(self, temperature: float) -> None:
        """Set the target temperature for the heated lid, in °C.

        Args:
            temperature: A value between 37 and 110, representing the target
                temperature in °C.

        !!! note
            The Thermocycler will proceed to the next command immediately after
            `temperature` is reached.
        """
        self._core.set_target_lid_temperature(celsius=temperature)
        self._core.wait_for_lid_temperature()

    @publish(command=cmds.thermocycler_start_set_lid_temperature)
    @requires_version(2, 27)
    def start_set_lid_temperature(self, temperature: float) -> Task:
        """Sets a target temperature to heat the Thermocycler Module's lid, in °C.

        Returns a [`Task`][opentrons.protocol_api.Task] object that represents concurrent heating.
        Pass the task object to [`wait_for_tasks()`][opentrons.protocol_api.ProtocolContext.wait_for_tasks]
        to wait for the lid to reach the target temperature.

        Args:
            temperature: A value between 37 and 110, representing the target
                temperature in °C.

        """
        task = self._core.start_set_target_lid_temperature(celsius=temperature)
        return Task(api_version=self._api_version, core=task)

    @publish(command=cmds.thermocycler_execute_profile)
    @requires_version(2, 0)
    def execute_profile(
        self,
        steps: List[ThermocyclerStep],
        repetitions: int,
        block_max_volume: Optional[float] = None,
    ) -> None:
        """
        Execute a Thermocycler profile, defined as a cycle of
        `steps`, for a given number of `repetitions`.

        Args:
            steps: List of steps that make up a single cycle.
                Each list item should be a dictionary that maps to the parameters
                of the [`set_block_temperature()`][opentrons.protocol_api.ThermocyclerContext.set_block_temperature]
                method. The dictionary's keys must be `temperature` and one or both of
                `hold_time_seconds` and `hold_time_minutes`.
            repetitions: The number of times to repeat the cycled steps.
            block_max_volume: The greatest volume of liquid contained in any
                individual well of the loaded labware, in µL. If not specified,
                the default is 25 µL.

        *Changed in version 2.21:* Fixed run log listing number of steps instead of
        number of repetitions.
        """
        repetitions = validation.ensure_thermocycler_repetition_count(repetitions)
        validated_steps = validation.ensure_thermocycler_profile_steps(steps)
        self._core.execute_profile(
            steps=validated_steps,
            repetitions=repetitions,
            block_max_volume=block_max_volume,
        )

    @publish(command=cmds.thermocycler_start_execute_profile)
    @requires_version(2, 27)
    def start_execute_profile(
        self,
        steps: List[ThermocyclerStep],
        repetitions: int,
        block_max_volume: Optional[float] = None,
    ) -> Task:
        """
        Starts a defined Thermocycler Module profile and returns a [`Task`][opentrons.protocol_api.Task] representing its concurrent execution.
        Profile is defined as a cycle of `steps`, for a given number of `repetitions`.

        Pass the task object to [`ProtocolContext.wait_for_tasks()`][opentrons.protocol_api.ProtocolContext.wait_for_tasks]
        to wait for the profile to complete.

        Args:
            steps: List of steps that make up a single cycle.
                Each list item should be a dictionary that maps to the parameters
                of the [`set_block_temperature()`][opentrons.protocol_api.ThermocyclerContext.set_block_temperature]
                method. The dictionary's keys must be `temperature` and one or both of
                `hold_time_seconds` and `hold_time_minutes`.
            repetitions: The number of times to repeat the cycled steps.
            block_max_volume: The greatest volume of liquid contained in any
                individual well of the loaded labware, in µL. If not specified, the
                default is 25 µL.
        """
        repetitions = validation.ensure_thermocycler_repetition_count(repetitions)
        validated_steps = validation.ensure_thermocycler_profile_steps(steps)
        task = self._core.start_execute_profile(
            steps=validated_steps,
            repetitions=repetitions,
            block_max_volume=block_max_volume,
        )
        return Task(api_version=self._api_version, core=task)

    @publish(command=cmds.thermocycler_deactivate_lid)
    @requires_version(2, 0)
    def deactivate_lid(self) -> None:
        """Turn off the lid heater."""
        self._core.deactivate_lid()

    @publish(command=cmds.thermocycler_deactivate_block)
    @requires_version(2, 0)
    def deactivate_block(self) -> None:
        """Turn off the well block temperature controller."""
        self._core.deactivate_block()

    @publish(command=cmds.thermocycler_deactivate)
    @requires_version(2, 0)
    def deactivate(self) -> None:
        """Turn off both the well block temperature controller and the lid heater."""
        self._core.deactivate()

    @property
    @requires_version(2, 0)
    def lid_position(self) -> Optional[str]:
        """One of these possible lid statuses:

        - `closed`: The lid is closed.
        - `in_between`: The lid is neither open nor closed.
        - `open`: The lid is open.
        - `unknown`: The lid position can't be determined.
        """
        status = self._core.get_lid_position()
        return status.value if status is not None else None

    @property
    @requires_version(2, 0)
    def block_temperature_status(self) -> str:
        """One of five possible temperature statuses:

        - `holding at target`: The block has reached its target temperature
          and is actively maintaining that temperature.
        - `cooling`: The block is cooling to a target temperature.
        - `heating`: The block is heating to a target temperature.
        - `idle`: The block is not currently heating or cooling.
        - `error`: The temperature status can't be determined.
        """
        return self._core.get_block_temperature_status().value

    @property
    @requires_version(2, 0)
    def lid_temperature_status(self) -> Optional[str]:
        """One of five possible temperature statuses:

        - `holding at target`: The lid has reached its target temperature
          and is actively maintaining that temperature.
        - `cooling`: The lid has previously heated and is now passively cooling.
            *The Thermocycler lid does not have active cooling.*
        - `heating`: The lid is heating to a target temperature.
        - `idle`: The lid has not heated since the beginning of the protocol.
        - `error`: The temperature status can't be determined.
        """
        status = self._core.get_lid_temperature_status()
        return status.value if status is not None else None

    @property
    @requires_version(2, 0)
    def block_temperature(self) -> Optional[float]:
        """The current temperature of the well block in °C."""
        return self._core.get_block_temperature()

    @property
    @requires_version(2, 0)
    def block_target_temperature(self) -> Optional[float]:
        """The target temperature of the well block in °C."""
        return self._core.get_block_target_temperature()

    @property
    @requires_version(2, 0)
    def lid_temperature(self) -> Optional[float]:
        """The current temperature of the lid in °C."""
        return self._core.get_lid_temperature()

    @property
    @requires_version(2, 0)
    def lid_target_temperature(self) -> Optional[float]:
        """The target temperature of the lid in °C."""
        return self._core.get_lid_target_temperature()

    @property
    @requires_version(2, 0)
    def ramp_rate(self) -> Optional[float]:
        """The current ramp rate in °C/s."""
        return self._core.get_ramp_rate()

    @property
    @requires_version(2, 0)
    def hold_time(self) -> Optional[float]:
        """Remaining hold time in seconds."""
        return self._core.get_hold_time()

    @property
    @requires_version(2, 0)
    def total_cycle_count(self) -> Optional[int]:
        """Number of repetitions for current set cycle"""
        return self._core.get_total_cycle_count()

    @property
    @requires_version(2, 0)
    def current_cycle_index(self) -> Optional[int]:
        """Index of the current set cycle repetition"""
        return self._core.get_current_cycle_index()

    @property
    @requires_version(2, 0)
    def total_step_count(self) -> Optional[int]:
        """Number of steps within the current cycle"""
        return self._core.get_total_step_count()

    @property
    @requires_version(2, 0)
    def current_step_index(self) -> Optional[int]:
        """Index of the current step within the current cycle"""
        return self._core.get_current_step_index()

    def _get_current_labware_max_vol(self) -> Optional[float]:
        max_vol: Optional[float] = None
        if self.labware is not None:
            for well in self.labware.wells():
                if well.has_tracked_liquid():
                    # make sure that max vol is a float first if we have liquid
                    max_vol = 0.0 if max_vol is None else max_vol
                    well_vol = well.current_liquid_volume()
                    # ignore simulated probe results
                    if isinstance(well_vol, float):
                        max_vol = max(max_vol, well_vol)
                    if max_vol > BLOCK_VOL_MAX:
                        max_vol = BLOCK_VOL_MAX
                    elif max_vol < BLOCK_VOL_MIN:
                        max_vol = BLOCK_VOL_MIN
        return max_vol


class HeaterShakerContext(ModuleContext):
    """An object representing a connected Heater-Shaker Module.

    It should not be instantiated directly; instead, it should be
    created through [`ProtocolContext.load_module()`][opentrons.protocol_api.ProtocolContext.load_module].

    *New in version 2.13*
    """

    _core: HeaterShakerCore

    @property
    @requires_version(2, 14)
    def serial_number(self) -> str:
        """Get the module's unique hardware serial number."""
        return self._core.get_serial_number()

    @property
    @requires_version(2, 13)
    def target_temperature(self) -> Optional[float]:
        """The target temperature of the Heater-Shaker's plate in °C.

        Returns `None` if no target has been set.
        """
        return self._core.get_target_temperature()

    @property
    @requires_version(2, 13)
    def current_temperature(self) -> float:
        """The current temperature of the Heater-Shaker's plate in °C.

        Returns `23` in simulation if no target temperature has been set.
        """
        return self._core.get_current_temperature()

    @property
    @requires_version(2, 13)
    def current_speed(self) -> int:
        """The current speed of the Heater-Shaker's plate in rpm."""
        return self._core.get_current_speed()

    @property
    @requires_version(2, 13)
    def target_speed(self) -> Optional[int]:
        """Target speed of the Heater-Shaker's plate in rpm."""
        return self._core.get_target_speed()

    @property
    @requires_version(2, 13)
    def temperature_status(self) -> str:
        """
        One of five possible temperature statuses:

        - `holding at target`: The module has reached its target temperature
          and is actively maintaining that temperature.
        - `cooling`: The module has previously heated and is now passively cooling.
          *The Heater-Shaker does not have active cooling.*
        - `heating`: The module is heating to a target temperature.
        - `idle`: The module has not heated since the beginning of the protocol.
        - `error`: The temperature status can't be determined.
        """
        return self._core.get_temperature_status().value

    @property
    @requires_version(2, 13)
    def speed_status(self) -> str:
        """One of five possible shaking statuses:

        - `holding at target`: The module has reached its target shake speed
          and is actively maintaining that speed.
        - `speeding up`: The module is increasing its shake speed towards a target.
        - `slowing down`: The module was previously shaking at a faster speed
          and is currently reducing its speed to a lower target or to deactivate.
        - `idle`: The module is not shaking.
        - `error`: The shaking status can't be determined.
        """
        return self._core.get_speed_status().value

    @property
    @requires_version(2, 13)
    def labware_latch_status(self) -> str:
        """One of six possible latch statuses:

        - `opening`: The latch is currently opening (in motion).
        - `idle_open`: The latch is open and not moving.
        - `closing`: The latch is currently closing (in motion).
        - `idle_closed`: The latch is closed and not moving.
        - `idle_unknown`: The default status upon reset, regardless of physical latch position.
          Use [`close_labware_latch()`][opentrons.protocol_api.HeaterShakerContext.close_labware_latch] before other commands
          requiring confirmation that the latch is closed.
        - `unknown`: The latch status can't be determined.
        """
        return self._core.get_labware_latch_status().value

    @requires_version(2, 13)
    def set_and_wait_for_temperature(self, celsius: float) -> None:
        """Set a target temperature and wait until the module reaches the target.

        No other protocol commands will execute while waiting for the temperature.

        *Changed in version 2.25:* Removed the minimum temperature limit of 37 °C.
        Note that temperatures under ambient are not achievable.

        Args:
            celsius: A value under 95, representing the target temperature in °C.
                Values are automatically truncated to two decimal places, and the
                Heater-Shaker module has a temperature accuracy of ±0.5 °C.
        """
        self.set_target_temperature(celsius=celsius)
        self.wait_for_temperature()

    @requires_version(2, 13)
    @publish(command=cmds.heater_shaker_set_target_temperature)
    def set_target_temperature(self, celsius: float) -> Task:
        """
        Set target temperature and return immediately.

        Sets the Heater-Shaker's target temperature and returns immediately without
        waiting for the target to be reached. Allows the protocol to proceed while the module reaches the target temperature.


        Use [`wait_for_temperature()`][opentrons.protocol_api.HeaterShakerContext.wait_for_temperature]
        to delay protocol execution for API levels below 2.27.

        *Changed in version 2.25:* Removed the minimum temperature limit of 37 °C.
        Note that temperatures under ambient are not achievable.

        *Changed in version 2.27:* Returns a [`Task`][opentrons.protocol_api.Task] object that represents concurrent
        heating. Pass the task object to
        [`wait_for_tasks()`][opentrons.protocol_api.ProtocolContext.wait_for_tasks]
        to wait for the module to reach the target temperature.

        Args:
            celsius: A value under 95, representing the target temperature in °C.
                Values are automatically truncated to two decimal places, and the
                Heater-Shaker module has a temperature accuracy of ±0.5 °C.
        """
        validated_temp = validate_heater_shaker_temperature(
            celsius=celsius, api_version=self.api_version
        )
        task = self._core.set_target_temperature(celsius=validated_temp)
        if self._api_version >= APIVersion(2, 27):
            return Task(api_version=self._api_version, core=task)
        else:
            return cast(Task, None)

    @requires_version(2, 13)
    @publish(command=cmds.heater_shaker_wait_for_temperature)
    def wait_for_temperature(self) -> None:
        """Delays protocol execution until the Heater-Shaker has reached its target
        temperature.

        Raises an error if no target temperature was previously set.
        """
        self._core.wait_for_target_temperature()

    @requires_version(2, 13)
    @publish(command=cmds.heater_shaker_set_and_wait_for_shake_speed)
    def set_and_wait_for_shake_speed(self, rpm: int) -> None:
        """Sets the Heater-Shaker's shake speed in RPM and returns a [`Task`][opentrons.protocol_api.Task] that represents concurrent shaking.

        !!! note
            Before shaking, this command retracts pipettes upward if they are adjacent
            to the Heater-Shaker Module.

        Args:
            rpm: A value between 200 and 3000, representing the target shake speed in
                revolutions per minute.
        """
        validated_speed = validate_heater_shaker_speed(rpm=rpm)
        self._core.set_and_wait_for_shake_speed(rpm=validated_speed)

    @requires_version(2, 27)
    @publish(command=cmds.heater_shaker_set_shake_speed)
    def set_shake_speed(self, rpm: int) -> Task:
        """Sets the Heater-Shaker's shake speed in RPM and returns a :py:class:`Task` that represents concurrent shaking.

        !!! note
            Before shaking, this command will retract the pipettes upward if they are
            parked adjacent to the Heater-Shaker.

        Args:
            rpm: A value between 200 and 3000, representing the target shake speed in
                revolutions per minute.
        """
        validated_speed = validate_heater_shaker_speed(rpm=rpm)
        task = self._core.set_shake_speed(rpm=validated_speed)
        return Task(api_version=self._api_version, core=task)

    @requires_version(2, 13)
    @publish(command=cmds.heater_shaker_open_labware_latch)
    def open_labware_latch(self) -> None:
        """Open the Heater-Shaker's labware latch.

        The labware latch needs to be closed before:

        - Shaking
        - Pipetting to or from the labware on the Heater-Shaker
        - Pipetting to or from labware to the left or right of the Heater-Shaker

        Attempting to open the latch while the Heater-Shaker is shaking will raise an error.

        !!! note
            Before opening the latch, this command will retract the pipettes upward
            if they are parked adjacent to the left or right of the Heater-Shaker.
        """
        self._core.open_labware_latch()

    @requires_version(2, 13)
    @publish(command=cmds.heater_shaker_close_labware_latch)
    def close_labware_latch(self) -> None:
        """Closes the labware latch.

        The labware latch needs to be closed using this method before sending a shake command,
        even if the latch was manually closed before starting the protocol.
        """
        self._core.close_labware_latch()

    @requires_version(2, 13)
    @publish(command=cmds.heater_shaker_deactivate_shaker)
    def deactivate_shaker(self) -> None:
        """Stops shaking.

        Decelerating to 0 rpm typically only takes a few seconds.
        """
        self._core.deactivate_shaker()

    @requires_version(2, 13)
    @publish(command=cmds.heater_shaker_deactivate_heater)
    def deactivate_heater(self) -> None:
        """Stops heating.

        The module will passively cool to room temperature.
        The Heater-Shaker does not have active cooling.
        """
        self._core.deactivate_heater()


class MagneticBlockContext(ModuleContext):
    """
    An object representing a Magnetic Block.

    It should not be instantiated directly; instead, it should be
    created through [`ProtocolContext.load_module()`][opentrons.protocol_api.ProtocolContext.load_module].

    *New in version 2.15*
    """

    _core: MagneticBlockCore


class AbsorbanceReaderContext(ModuleContext):
    """
    An object representing a connected Absorbance Plate Reader Module.

    It should not be instantiated directly; instead, it should be
    created through [`ProtocolContext.load_module()`][opentrons.protocol_api.ProtocolContext.load_module].

    *New in version 2.21*
    """

    _core: AbsorbanceReaderCore

    @property
    @requires_version(2, 21)
    def serial_number(self) -> str:
        """Get the module's unique hardware serial number."""
        return self._core.get_serial_number()

    @requires_version(2, 21)
    def close_lid(self) -> None:
        """Use the Flex Gripper to close the lid of the Absorbance Plate Reader.

        You must call this method before initializing the reader, even if the reader was
        in the closed position at the start of the protocol.
        """
        self._core.close_lid()

    @requires_version(2, 21)
    def open_lid(self) -> None:
        """Use the Flex Gripper to open the lid of the Absorbance Plate Reader."""
        self._core.open_lid()

    @requires_version(2, 21)
    def is_lid_on(self) -> bool:
        """Return `True` if the Absorbance Plate Reader's lid is currently closed."""
        return self._core.is_lid_on()

    @requires_version(2, 21)
    def initialize(
        self,
        mode: ABSMeasureMode,
        wavelengths: List[int],
        reference_wavelength: Optional[int] = None,
    ) -> None:
        """Prepare the Absorbance Plate Reader to read a plate.

        See [Initialization](../modules/absorbance-plate-reader.md#initialization) for examples.

        Args:
            mode: Either `"single"` or `"multi"`.

                - In single measurement mode, [`read()`][opentrons.protocol_api.AbsorbanceReaderContext.read]
                uses one sample wavelength and an optional reference wavelength.
                - In multiple measurement mode, [`read()`][opentrons.protocol_api.AbsorbanceReaderContext.read]
                uses a list of up to six sample wavelengths.
            wavelengths: A list of wavelengths, in nm, to measure.

                - In the default hardware configuration, each wavelength must be one of
                `450` (blue), `562` (green), `600` (orange), or `650` (red). In custom
                hardware configurations, the module may accept other integers between
                `350` and `1000`.
                - The list must contain only one item when initializing a single measurement.
                - The list can contain one to six items when initializing a multiple measurement.
            reference_wavelength: An optional reference wavelength, in nm. If provided,
                [`read()`][opentrons.protocol_api.AbsorbanceReaderContext.read] will read at the
                reference wavelength and then subtract the reference wavelength values from the
                measurement wavelength values. Can only be used with single measurements.
        """
        self._core.initialize(
            mode, wavelengths, reference_wavelength=reference_wavelength
        )

    @requires_version(2, 21)
    def read(
        self, export_filename: Optional[str] = None
    ) -> Dict[int, Dict[str, float]]:
        """
        Read a plate on the Absorbance Plate Reader.

        This method always returns a dictionary of measurement data. It optionally will
        save a CSV file of the results to the Flex filesystem, which you can access from
        the Recent Protocol Runs screen in the Opentrons App. These files are *only* saved
        if you specify `export_filename`.

        In simulation, the values for each well key in the dictionary are set to zero, and
        no files are written.

        !!! note
            Avoid divide-by-zero errors when simulating and using the results of this
            method later in the protocol. If you divide by any of the measurement
            values, use [`is_simulating()`][opentrons.protocol_api.ProtocolContext.is_simulating]
            to use alternate dummy data or skip the division step.

        Args:
            export_filename: An optional file basename. If provided, this method
                will write a CSV file for each measurement in the read operation. File
                names will use the value of this parameter, the measurement wavelength
                supplied in [`initialize()`][opentrons.protocol_api.AbsorbanceReaderContext.initialize],
                and a `.csv` extension. For example, when reading at wavelengths 450 and 562
                with `export_filename="my_data"`, there will be two output files:
                `my_data_450.csv` and `my_data_562.csv`.

                See [CSV Data](../modules/absorbance-plate-reader.md#csv-data)
                for information on working with these CSV files.

        Returns:
            A dictionary of wavelengths to dictionary of values ordered by well name.
        """
        return self._core.read(filename=export_filename)


class FlexStackerContext(ModuleContext):
    """An object representing a connected Flex Stacker module.

    It should not be instantiated directly; instead, it should be
    created through [`ProtocolContext.load_module()`][opentrons.protocol_api.ProtocolContext.load_module].

    *New in version 2.25*
    """

    _core: FlexStackerCore

    @property
    @requires_version(2, 25)
    def serial_number(self) -> str:
        """Get the module's unique hardware serial number."""
        return self._core.get_serial_number()

    @requires_version(2, 25)
    @publish(command=cmds.flex_stacker_retrieve)
    def retrieve(self) -> Labware:
        """
        Retrieve a labware from the Flex Stacker and move it onto the shuttle.

        The Stacker will retrieve the bottom-most labware in the stack.

        Returns:
            The retrieved [`Labware`][opentrons.protocol_api.labware.Labware] object. This will always
                be the main labware, even if the Flex Stacker contains labware on an adapter. To get the
                adapter object, call [`Labware.parent`][opentrons.protocol_api.labware.Labware.parent] on
                the returned labware.
        """
        labware_core = self._core.retrieve()

        return self._core_map.get_or_add(
            labware_core,
            Labware._builder_for_core_map(
                self._api_version, self._protocol_core, self._core_map
            ),
        )

    @requires_version(2, 25)
    @publish(command=cmds.flex_stacker_store)
    def store(self) -> None:
        """Move a labware currently on the Flex Stacker shuttle into the Flex Stacker.

        The labware must be the same type the Stacker is configured to store using
        [`set_stored_labware()`][opentrons.protocol_api.FlexStackerContext.set_stored_labware].
        If labware is currently stacked inside the module, this method moves the new
        labware to the bottom-most position of the stack.
        """
        self._core.store()

    def _labware_to_cores(self, labware: Sequence[Labware]) -> list[LabwareCore]:
        return [labware._core for labware in labware]

    def _cores_to_labware(self, cores: Sequence[LabwareCore]) -> list[Labware]:
        def _convert() -> Iterator[Labware]:
            for core in cores:
                yield self._core_map.get_or_add(
                    core,
                    Labware._builder_for_core_map(
                        self._api_version, self._protocol_core, self._core_map
                    ),
                )

        return list(_convert())

    @requires_version(2, 25)
    def get_max_storable_labware_from_list(
        self,
        labware: list[Labware],
        stacking_offset_z: float | None = None,
    ) -> list[Labware]:
        """
        Limit a list of labware instances to the number that can be stored in a Flex Stacker.
        Items will be taken from the head of the list.

        A Flex Stacker has a limited amount of internal space and computes the number of labware
        (or labware with lids or adapters) that it can store based on the `z` heights of the labware
        and the amount they overlap when stacked. To calculate how many of a given
        labware the Stacker can store, the labware type must be specified.

        Provide a list of labware to this function to return the maximum number of labware of the given type that the
        Stacker can store. The returned list is guaranteed to be suitable
        for passing to [`set_stored_labware_items()`][opentrons.protocol_api.FlexStackerContext.set_stored_labware_items].

        This function limits the list of labware based on the overall maximum number the Stacker
        can hold and will not change as labware is added or removed. To limit a list of labware to
        the amount that will currently fit in the Flex Stacker, use
        [`get_current_storable_labware_from_list()`][opentrons.protocol_api.FlexStackerContext.get_current_storable_labware_from_list].

        !!! note
            If a `z` stacking offset is provided, be sure to specify the same value when
            configuring the Flex Stacker with
            [`set_stored_labware_items()`][opentrons.protocol_api.FlexStackerContext.set_stored_labware_items].

            See [`set_stored_labware_items()`][opentrons.protocol_api.FlexStackerContext.set_stored_labware_items]
            for more details on stacking offset.
        """
        return self._cores_to_labware(
            self._core.get_max_storable_labware_from_list(
                self._labware_to_cores(labware), stacking_offset_z
            ),
        )

    @requires_version(2, 25)
    def get_current_storable_labware_from_list(
        self, labware: list[Labware]
    ) -> list[Labware]:
        """
        Limit a list of labware instances to the number that the Flex Stacker currently has space for,
        based on the labware that is already stored in the Flex Stacker. Items will be taken from the
        head of the list.

        A Flex Stacker has a limited amount of internal space and computes the number of labware that
        it can store based on the `z` height of the labware and the amount they overlap when stacked.

        !!! note
            The number of elements in the returned list will change as labware is added or removed from
            the Flex Stacker. To get a list limited to the overall maximum number of labware the Flex
            Stacker can store, use
            [`get_max_storable_labware_from_list()`][opentrons.protocol_api.FlexStackerContext.get_max_storable_labware_from_list].

        Args:
            labware: A list of labware to limit. The returned list takes from the front of the provided
                list, and it is guaranteed to be suitable for passing to
                [`fill_items()`][opentrons.protocol_api.FlexStackerContext.fill_items].
        """
        return self._cores_to_labware(
            self._core.get_current_storable_labware_from_list(
                self._labware_to_cores(labware)
            )
        )

    @requires_version(2, 25)
    def get_max_storable_labware(self) -> int:
        """Get the maximum number of labware that the Flex Stacker can store.

        Use this function to return the total number of labware that the Flex Stacker can store. A Stacker has a limited
        amount of internal space and calculates the total number of labware that can be stored based on the `z` height
        of the labware and the amount they overlap when stacked.

        The total number is calculated based on the labware definition for the type of labware the Stacker is currently
        configured to store using [`set_stored_labware()`][opentrons.protocol_api.FlexStackerContext.set_stored_labware].
        This number is the overall maximum and will not change as labware is added or removed. To get the number of
        labware that can be stored in the Flex Stacker based on its current conditions, use
        [`get_current_storable_labware()`][opentrons.protocol_api.FlexStackerContext.get_current_storable_labware].
        """
        return self._core.get_max_storable_labware()

    @requires_version(2, 25)
    def get_current_storable_labware(self) -> int:
        """Get the number of labware that the Flex Stacker currently has space for.

        Use this function to return the total number of labware that the Flex Stacker
        can store. A Stacker has a limited amount of internal space and calculates the
        number of labware that can be stored based on the `z` height of the labware
        and the amount they overlap when stacked.

        The number is calculated based on the labware definition for the type of
        labware the Stacker is currently configured to store using
        [`set_stored_labware()`][opentrons.protocol_api.FlexStackerContext.set_stored_labware].
        This function returns a number based on the current storage conditions of the
        Stacker, and will change as labware is added or removed. To get the overall
        maximum number of labware the Flex Stacker can store, use
        [`get_max_storable_labware()`][opentrons.protocol_api.FlexStackerContext.get_max_storable_labware].
        """
        return self._core.get_current_storable_labware()

    @requires_version(2, 25)
    def set_stored_labware_items(
        self,
        labware: list[Labware],
        stacking_offset_z: float | None = None,
    ) -> None:
        """
        Configure the labware the Flex Stacker will store during a protocol by
        providing an initial list of stored labware objects. The start of the list
        represents the bottom of the Stacker, and the end of the list represents the
        top of the Stacker.

        The kind of labware stored by the Flex Stacker will be calculated from the
        list of labware specified here. You can use this to store labware objects
        that you have already created so that, for instance, you can set their
        liquid state or nicknames.

        Args:
            labware: A list of labware to load into the Stacker.

                - The list must have at least one element.
                - All labware must be loaded [`OFF_DECK`][opentrons.protocol_api.OFF_DECK].
                - All labware must be of the same kind. If any of them have lids, they
                must all have lids, and the lids must be the same. If any of them are
                on adapters, they all must be on adapters, and the adapters must be
                the same. All lids and adapters must be compatible with the Stacker.
                - The number of labware objects must fit in the Stacker physically. To
                make sure the labware will fit, use the return value of
                [`get_max_storable_labware_from_list()`][opentrons.protocol_api.FlexStackerContext.get_max_storable_labware_from_list].

            stacking_offset_z: Stacking `z` offset in mm of stored labware. If
                specified, this overrides the calculated value from labware
                definitions.

        !!! note
            The stacking offset is the amount of vertical overlap (in mm) between
            the bottom side of a labware unit and the top side of the unit below.
            This offset is used to determine how many units can fit in the stacker
            and calculates the `z` position of the shuttle when retrieving or
            storing labware.

            There are four possible stacking configurations, each with a different
            method of calculating the stacking offset:

            - Bare labware: labware (bottom side) overlaps with the top side of
              the labware below.
            - Labware on adapter: the adapter (bottom side) of the upper labware
              unit overlaps with the top side of the labware below.
            - Labware with lid: the labware (bottom side) of the upper unit
              overlaps with the lid (top side) of the unit below.
            - Labware with lid and adapter: the adapter (bottom side) of the
              upper unit overlaps with the lid (top side) of the unit below.
        """
        self._core.set_stored_labware_items(
            self._labware_to_cores(labware),
            stacking_offset_z=stacking_offset_z,
        )

    @requires_version(2, 25)
    @publish(command=cmds.flex_stacker_set_stored_labware)
    def set_stored_labware(
        self,
        load_name: str,
        namespace: str | None = None,
        version: int | None = None,
        adapter: str | None = None,
        lid: str | None = None,
        count: int | None = None,
        stacking_offset_z: float | None = None,
        *,
        adapter_namespace: str | None = None,
        adapter_version: int | None = None,
        lid_namespace: str | None = None,
        lid_version: int | None = None,
    ) -> None:
        """
        Configure the type and starting quantity of labware the Flex Stacker will store during a protocol. This is the
        only type of labware you'll be able to store in the Stacker until it's reconfigured.

        You must use this method to load a labware stack stored inside the Stacker before you're able to
        [`retrieve()`][opentrons.protocol_api.FlexStackerContext.retrieve] or
        [`store()`][opentrons.protocol_api.FlexStackerContext.store] additional labware.

        Args:
            load_name (str): A string to use for looking up a labware definition.
                You can find the `load_name` for any Opentrons-verified labware on the
                [Labware Library](https://labware.opentrons.com).

            namespace (str): The namespace that the labware definition belongs to.
                If unspecified, the API will automatically search two namespaces:

                - `"opentrons"`, to load standard Opentrons labware definitions.
                - `"custom_beta"`, to load custom labware definitions created with the
                [Custom Labware Creator](https://labware.opentrons.com/create).

                You might need to specify an explicit `namespace` if you have a custom
                definition whose `load_name` is the same as an Opentrons-verified
                definition, and you want to explicitly choose one or the other.

            version (int): The version of the labware definition. You should normally
                leave this unspecified to let the method choose a version
                automatically.

            adapter (str): An adapter to load the labware on top of. Accepts the same
                values as the `load_name` parameter of
                [`load_adapter()`][opentrons.protocol_api.ProtocolContext.load_adapter].

            adapter_namespace (str): Applies to `adapter` the same way that `namespace`
                applies to `load_name`.

                *Changed in version 2.26:* `adapter_namespace` may now be specified explicitly.
                When you've specified `namespace` for `load_name` but not `adapter_namespace`,
                `adapter_namespace` now independently follows the same search rules described
                in `namespace`. Formerly, it took the exact `namespace` value.

            adapter_version (int): Applies to `adapter` the same way that `version`
                applies to `load_name`.

                *Changed in version 2.26:* `adapter_version` may now be specified explicitly.
                When unspecified, improved search rules prevent selecting a version that does
                not exist.

            lid (str): A lid to load on top of the main labware. Accepts the same
                values as the `load_name` parameter of
                [`load_lid_stack()`][opentrons.protocol_api.ProtocolContext.load_lid_stack].
                The lid will use the same namespace as the labware, and the API will
                choose the lid's version automatically.

                *New in version 2.23*

            lid_namespace (str): The namespace of the lid being loaded.
                Applies to lid the same way that namespace applies to load_name.

                *Changed in version 2.26:* `lid_namespace` may now be specified explicitly.
                When you've specified `namespace` for `load_name` but not `lid_namespace`,
                `lid_namespace` now independently follows the same search rules
                described in `namespace`. Formerly, it took the exact `namespace` value.

            lid_version (int): Applies to `lid` the same way that `version`
                applies to `load_name`.

                *Changed in version 2.26:* `lid_version` may now be specified explicitly.
                When unspecified, improved search rules prevent selecting a version that does
                not exist.

            count (int): The number of labware that the Flex Stacker should store. If not specified,
                this will be the maximum amount of this kind of labware that the Flex Stacker is
                capable of storing.

            stacking_offset_z (float): Stacking `z` offset in mm of stored labware. If specified, this
                overrides the calculated value in the labware definition.

                !!! note
                    The stacking offset is the amount of vertical overlap (in mm) between the bottom side of a
                    labware unit and the top side of the unit below. This offset is used to determine how many
                    units can fit in the Stacker and calculates the `z` position of the shuttle when retrieving
                    or storing labware.

                    There are four possible stacking configurations, each with a different method of calculating
                    the stacking offset:

                    - Bare labware: labware (bottom side) overlaps with the top side of the labware below.
                    - Labware on adapter: the adapter (bottom side) of the upper labware unit overlaps with the
                    top side of the labware below.
                    - Labware with lid: the labware (bottom side) of the upper unit overlaps with the lid (top
                    side) of the unit below.
                    - Labware with lid and adapter: the adapter (bottom side) of the upper unit overlaps with
                    the lid (top side) of the unit below.
        """

        if self._api_version < validation.NAMESPACE_VERSION_ADAPTER_LID_VERSION_GATE:
            if adapter_namespace is not None:
                raise APIVersionError(
                    api_element="The `adapter_namespace` parameter",
                    until_version=str(
                        validation.NAMESPACE_VERSION_ADAPTER_LID_VERSION_GATE
                    ),
                    current_version=str(self._api_version),
                )
            if adapter_version is not None:
                raise APIVersionError(
                    api_element="The `adapter_version` parameter",
                    until_version=str(
                        validation.NAMESPACE_VERSION_ADAPTER_LID_VERSION_GATE
                    ),
                    current_version=str(self._api_version),
                )
            if lid_namespace is not None:
                raise APIVersionError(
                    api_element="The `lid_namespace` parameter",
                    until_version=str(
                        validation.NAMESPACE_VERSION_ADAPTER_LID_VERSION_GATE
                    ),
                    current_version=str(self._api_version),
                )
            if lid_version is not None:
                raise APIVersionError(
                    api_element="The `lid_version` parameter",
                    until_version=str(
                        validation.NAMESPACE_VERSION_ADAPTER_LID_VERSION_GATE
                    ),
                    current_version=str(self._api_version),
                )

        if self._api_version < validation.NAMESPACE_VERSION_ADAPTER_LID_VERSION_GATE:
            checked_adapter_namespace = namespace
            checked_adapter_version = version
            checked_lid_namespace = namespace
            checked_lid_version = version
        else:
            checked_adapter_namespace = adapter_namespace
            checked_adapter_version = adapter_version
            checked_lid_namespace = lid_namespace
            checked_lid_version = lid_version

        self._core.set_stored_labware(
            main_load_name=load_name,
            main_namespace=namespace,
            main_version=version,
            lid_load_name=lid,
            lid_namespace=checked_lid_namespace,
            lid_version=checked_lid_version,
            adapter_load_name=adapter,
            adapter_namespace=checked_adapter_namespace,
            adapter_version=checked_adapter_version,
            count=count,
            stacking_offset_z=stacking_offset_z,
        )

    @requires_version(2, 25)
    @publish(command=cmds.flex_stacker_fill)
    def fill(self, count: int | None = None, message: str | None = None) -> None:
        """Pause the protocol to add labware to the Flex Stacker.

        The labware must be the same type the Stacker is configured to store using
        [`set_stored_labware()`][opentrons.protocol_api.FlexStackerContext.set_stored_labware].
        If no labware type has been set, the API will raise an error.

        Args:
            count: The amount of labware the Flex Stacker should hold after this command
                is executed. If not specified, the Flex Stacker should be full after this
                command is executed.
            message: A message to display noting what kind of labware to fill the Stacker with.
        """
        self._core.fill(count, message)

    @requires_version(2, 25)
    def fill_items(self, labware: list[Labware], message: str | None = None) -> None:
        """Pause the protocol to add a specific list of labware to the Flex Stacker.

        Args:
            labware: The list of labware to add. The list must:

                - Contain at least one labware.
                - Have labware of the same kind previously passed to
                [`set_stored_labware_items()`][opentrons.protocol_api.FlexStackerContext.set_stored_labware_items]
                or loaded by [`set_stored_labware()`][opentrons.protocol_api.FlexStackerContext.set_stored_labware].
                - All labware should be loaded [`OFF_DECK`][opentrons.protocol_api.OFF_DECK].
            message: A message to display noting the labware to fill the Stacker with.
        """
        self._core.fill_items(self._labware_to_cores(labware), message)

    @requires_version(2, 25)
    @publish(command=cmds.flex_stacker_empty)
    def empty(self, message: str | None = None) -> None:
        """
        Pause the protocol to remove all labware stored in the Flex Stacker.

        This method sets the location of all labware currently in the stacker to
        [`OFF_DECK`][opentrons.protocol_api.OFF_DECK].

        Args:
            message: A message to display to note what should be removed from
                the Flex Stacker.
        """
        self._core.empty(
            message,
        )

    @requires_version(2, 25)
    def get_stored_labware(self) -> list[Labware]:
        """Get the list of labware currently stored inside the Stacker.

        This function returns a list of all labware stored in the Stacker based on
        the labware initially stored using
        [`set_stored_labware()`][opentrons.protocol_api.FlexStackerContext.set_stored_labware]
        and any labware added or removed during the protocol.

        The first element of the list occupies the bottom-most position in the
        labware stack and would be the labware retrieved by a call to
        [`retrieve()`][opentrons.protocol_api.FlexStackerContext.retrieve].
        """
        return self._cores_to_labware(self._core.get_stored_labware())


class VacuumModuleContext(ModuleContext):
    """An object representing a connected Vacuum Module.

    It should not be instantiated directly; instead, it should be
    created through :py:meth:`.ProtocolContext.load_module`.

    """

    _core: VacuumModuleCore

    @property
    @requires_version(2, 28)
    def serial_number(self) -> str:
        """Get the module's unique hardware serial number."""
        return self._core.get_serial_number()



from typing import List, Mapping, Tuple

import pytest

from opentrons.hardware_control.errors import OutOfBoundsMove
from opentrons.hardware_control.types import (
    Axis,
    CriticalPoint,
    MotionChecks,
)
from opentrons.hardware_control.util import AxisType, check_motion_bounds, plan_arc
from opentrons.types import Point


def check_arc_basic(arc: List[Point], from_pt: Point, to_pt: Point) -> None:
    """Check the tests that should always be true for different-well moves
    - we should always go only up, then only xy, then only down
    - we should have three moves
    """
    # first move should move only up
    assert arc[0]._replace(z=0) == from_pt._replace(z=0)
    # second move should move only in xy
    assert arc[0].z == arc[1].z
    # second-to-last move should always end at the dest in xy
    # so the last move is z-only
    assert arc[-2]._replace(z=0) == to_pt._replace(z=0)
    # final move should arrive precisely at the dest
    assert arc[-1] == to_pt
    # first move should be up
    assert arc[0].z >= from_pt.z
    # last move should be down
    assert arc[-2].z >= to_pt.z


@pytest.mark.parametrize(
    argnames=["from_pt", "to_pt", "z_height"],
    argvalues=[
        [Point(10, 20, 30), Point(10, 30, 30), 100],
        [Point(10, 20, 30), Point(10, 30, 40), 100],
        [Point(10, 20, 40), Point(10, 30, 30), 40],
        [Point(10, 20, 30), Point(10, 30, 40), 40],
    ],
)
def test_basic_arcs(from_pt: Point, to_pt: Point, z_height: float) -> None:
    check_arc_basic([a[0] for a in plan_arc(from_pt, to_pt, z_height)], from_pt, to_pt)


def test_arc_with_waypoint() -> None:
    from_pt = Point(20, 20, 40)
    to_pt = Point(0, 0, 10)
    arc = plan_arc(from_pt, to_pt, 50, extra_waypoints=[(5, 10), (20, 30)])
    check_arc_basic([a[0] for a in arc], from_pt, to_pt)
    assert arc[1][0].x == 5
    assert arc[1][0].y == 10
    assert arc[1][0].z == 50
    assert arc[2][0].x == 20
    assert arc[2][0].y == 30
    assert arc[2][0].z == 50


def test_cp_blending() -> None:
    from_pt = Point(10, 10, 10)
    to_pt = Point(0, 0, 10)
    arc = plan_arc(from_pt, to_pt, 50, None, CriticalPoint.XY_CENTER)
    check_arc_basic([a[0] for a in arc], from_pt, to_pt)
    assert arc[0][1] is None
    assert arc[1][1] is CriticalPoint.XY_CENTER
    assert arc[2][1] is CriticalPoint.XY_CENTER

    arc2 = plan_arc(from_pt, to_pt, 50, CriticalPoint.TIP, None)
    check_arc_basic([a[0] for a in arc], from_pt, to_pt)
    assert arc2[0][1] == CriticalPoint.TIP
    assert arc2[1][1] is None
    assert arc2[2][1] is None


@pytest.mark.parametrize(
    "xformed,deck,bounds,check,catch,phrase",
    [
        # acceptable moves, iterating through checks
        ({Axis.X: 2}, {Axis.X: -15}, {Axis.X: (1, 4)}, MotionChecks.NONE, False, ""),
        ({Axis.X: 2}, {Axis.X: -10}, {Axis.X: (1, 4)}, MotionChecks.LOW, False, ""),
        ({Axis.X: 2}, {Axis.X: 0}, {Axis.X: (1, 4)}, MotionChecks.HIGH, False, ""),
        ({Axis.X: 2}, {Axis.X: 10}, {Axis.X: (1, 4)}, MotionChecks.BOTH, False, ""),
        # unacceptable low, with and without checking
        ({Axis.Z: 1}, {Axis.Z: 3}, {Axis.Z: (2, 4)}, MotionChecks.NONE, False, ""),
        ({Axis.Z: 1}, {Axis.Z: 3}, {Axis.Z: (2, 4)}, MotionChecks.LOW, True, "low"),
        ({Axis.Z: 1}, {Axis.Z: 3}, {Axis.Z: (2, 4)}, MotionChecks.HIGH, False, ""),
        ({Axis.Z: 1}, {Axis.Z: 3}, {Axis.Z: (2, 4)}, MotionChecks.BOTH, True, "low"),
        # unacceptable high, with and without checking
        ({Axis.Z: 5}, {Axis.Z: 3}, {Axis.Z: (2, 4)}, MotionChecks.NONE, False, ""),
        ({Axis.Z: 5}, {Axis.Z: 3}, {Axis.Z: (2, 4)}, MotionChecks.LOW, False, ""),
        ({Axis.Z: 5}, {Axis.Z: 3}, {Axis.Z: (2, 4)}, MotionChecks.HIGH, True, "high"),
        ({Axis.Z: 5}, {Axis.Z: 3}, {Axis.Z: (2, 4)}, MotionChecks.BOTH, True, "high"),
    ],
)
def test_check_motion_bounds(
    xformed: Mapping[AxisType, float],
    deck: Mapping[AxisType, float],
    bounds: Mapping[AxisType, Tuple[float, float]],
    check: MotionChecks,
    catch: bool,
    phrase: str,
) -> None:
    if catch:
        with pytest.raises(OutOfBoundsMove, match=phrase):
            check_motion_bounds(xformed, deck, bounds, check)
    else:
        check_motion_bounds(xformed, deck, bounds, check)



"""Test capture image command."""

from typing import Tuple

import mock
import pytest
from decoy import Decoy

from opentrons.protocol_engine.commands.capture_image import (
    CaptureImageImpl,
    CaptureImageParams,
    CaptureImageResult,
)
from opentrons.protocol_engine.commands.command import SuccessData
from opentrons.protocol_engine.errors import (
    CameraCaptureError,
    CameraDisabledError,
    CameraSettingsInvalidError,
    FileNameInvalidError,
)
from opentrons.protocol_engine.resources import CameraProvider, FileProvider
from opentrons.protocol_engine.resources.camera_provider import CameraSettings
from opentrons.protocol_engine.resources.file_provider import SPECIAL_CHARACTERS
from opentrons.protocol_engine.state import update_types
from opentrons.protocol_engine.state.state import StateView
from opentrons.protocol_engine.types import PreconditionTypes
from opentrons.system import camera, ffmpeg
from opentrons.system.camera import RESOLUTION_DEFAULT, ZOOM_DEFAULT, image_capture


@pytest.fixture(autouse=True)
def mock_ffmpeg_image_capture(decoy: Decoy, monkeypatch: pytest.MonkeyPatch) -> None:
    """Mock out the opentrons.ffmpeg image byte capture service."""
    monkeypatch.setattr(
        ffmpeg,
        "ffmpeg_capture_image_bytes",
        decoy.mock(func=ffmpeg.ffmpeg_capture_image_bytes),
    )


@pytest.fixture(autouse=True)
def mock_camera_stop_live_stream(decoy: Decoy, monkeypatch: pytest.MonkeyPatch) -> None:
    """Mock out the opentrons.camera live stream service stopper."""
    monkeypatch.setattr(
        camera,
        "stop_live_stream",
        decoy.mock(func=camera.stop_live_stream),
    )


@pytest.fixture(autouse=True)
def mock_camera_restart_live_stream(
    decoy: Decoy, monkeypatch: pytest.MonkeyPatch
) -> None:
    """Mock out the opentrons.camera live stream service restarter."""
    monkeypatch.setattr(
        camera,
        "restart_live_stream",
        decoy.mock(func=camera.restart_live_stream),
    )


@pytest.fixture
def camera_provider_image_capture() -> CameraProvider:
    """Creates a CameraProvider with image_capture hookups."""
    return CameraProvider(
        camera_settings_callback=None, image_capture_callback=image_capture
    )


@pytest.mark.parametrize(
    argnames=[
        "zoom",
        "contrast",
        "brightness",
        "saturation",
    ],
    argvalues=[
        [3.0, None, None, None],
        [-1.0, None, None, None],
        [None, 101.0, None, None],
        [None, -1.0, None, None],
        [None, None, 101.0, None],
        [None, None, -1.0, None],
        [None, None, None, 101.0],
        [None, None, None, -1.0],
    ],
)
async def test_raises_camera_settings_invalid_error(
    decoy: Decoy,
    state_view: StateView,
    file_provider: FileProvider,
    camera_provider_image_capture: CameraProvider,
    zoom: float | None,
    brightness: int | None,
    contrast: float | None,
    saturation: float | None,
) -> None:
    """It should raise CameraSettingsInvalidError when parameters would exceed limits."""
    subject = CaptureImageImpl(
        state_view=state_view,
        file_provider=file_provider,
        camera_provider=camera_provider_image_capture,
    )
    params = CaptureImageParams(
        fileName=None,
        zoom=zoom,
        brightness=brightness,
        contrast=contrast,
        saturation=saturation,
    )
    decoy.when(state_view.files.get_filecount()).then_return(0)

    with mock.patch("os.path.exists", mock.Mock(return_value=True)):
        with pytest.raises(CameraSettingsInvalidError):
            await subject.execute(params=params)


async def test_raises_camera_disabled_error(
    decoy: Decoy,
    state_view: StateView,
    file_provider: FileProvider,
    camera_provider: CameraProvider,
) -> None:
    """It should raise CameraDisabledError when the camera is not enabled in the boolean settings."""
    subject = CaptureImageImpl(
        state_view=state_view,
        file_provider=file_provider,
        camera_provider=camera_provider,
    )
    params = CaptureImageParams(fileName=None)
    decoy.when(state_view.files.get_filecount()).then_return(0)
    decoy.when(await camera_provider.get_camera_settings()).then_return(
        CameraSettings(
            cameraEnabled=False,
            liveStreamEnabled=False,
            errorRecoveryCameraEnabled=False,
        )
    )

    with pytest.raises(CameraDisabledError):
        await subject.execute(params=params)


async def test_raises_camera_camera_capture_error(
    decoy: Decoy,
    state_view: StateView,
    file_provider: FileProvider,
    camera_provider_image_capture: CameraProvider,
) -> None:
    """It should raise CameraCaptureError when a camera device doesn't exist."""
    subject = CaptureImageImpl(
        state_view=state_view,
        file_provider=file_provider,
        camera_provider=camera_provider_image_capture,
    )
    params = CaptureImageParams(fileName=None)
    decoy.when(state_view.files.get_filecount()).then_return(0)
    # Without decoying the video device this will return an error because the device doesn't exist
    with pytest.raises(CameraCaptureError):
        await subject.execute(params=params)


async def test_capture_image_returns_success(
    decoy: Decoy,
    state_view: StateView,
    file_provider: FileProvider,
    camera_provider_image_capture: CameraProvider,
) -> None:
    """It should return the successful result of an image capture with valid parameters."""
    subject = CaptureImageImpl(
        state_view=state_view,
        file_provider=file_provider,
        camera_provider=camera_provider_image_capture,
    )
    params = CaptureImageParams(fileName="coolpic")
    decoy.when(state_view.files.get_filecount()).then_return(0)

    with mock.patch("os.path.exists", mock.Mock(return_value=True)):
        result = await subject.execute(params=params)
        assert isinstance(result, SuccessData)


async def test_ensure_camera_used_precondition_set(
    decoy: Decoy,
    state_view: StateView,
    file_provider: FileProvider,
    camera_provider_image_capture: CameraProvider,
) -> None:
    """It should validate that the isCamerUsed precondition is set after an image is captured."""
    subject = CaptureImageImpl(
        state_view=state_view,
        file_provider=file_provider,
        camera_provider=camera_provider_image_capture,
    )
    params = CaptureImageParams(fileName="coolpic")
    decoy.when(state_view.files.get_filecount()).then_return(0)

    with mock.patch("os.path.exists", mock.Mock(return_value=True)):
        result = await subject.execute(params=params)
        assert isinstance(result, SuccessData)
        assert isinstance(
            result.state_update.precondition_update, update_types.PreconditionUpdate
        )
        assert result.state_update.precondition_update.preconditions == {
            PreconditionTypes.IS_CAMERA_USED: True
        }


@pytest.mark.parametrize(
    argnames=[
        "resolution",
        "result_resolution",
        "zoom",
        "result_zoom",
        "pan",
        "result_pan",
        "contrast",
        "result_contrast",
        "brightness",
        "result_brightness",
        "saturation",
        "result_saturation",
    ],
    argvalues=[
        [
            None,
            RESOLUTION_DEFAULT,
            None,
            ZOOM_DEFAULT,
            None,
            None,
            None,
            50,
            None,
            50,
            None,
            50,
        ],
        [
            (320, 240),
            (320, 240),
            1.5,
            1.5,
            (3, 4),
            (3, 4),
            25,
            25,
            10,
            10,
            75,
            75,
        ],
        [
            (320, 240),
            (320, 240),
            2.0,
            2.0,
            (3, 4),
            (3, 4),
            75,
            75,
            99,
            99,
            25,
            25,
        ],
        [
            (7680, 4320),
            (7680, 4320),
            1.0,
            1.0,
            (25, 45),
            (25, 45),
            21,
            21,
            11,
            11,
            59,
            59,
        ],
    ],
)
async def test_capture_image_returns_expected_params(
    decoy: Decoy,
    state_view: StateView,
    file_provider: FileProvider,
    camera_provider_image_capture: CameraProvider,
    resolution: Tuple[int, int] | None,
    zoom: float | None,
    pan: Tuple[int, int] | None,
    contrast: float | None,
    brightness: float | None,
    saturation: float | None,
    result_resolution: Tuple[int, int],
    result_zoom: float,
    result_pan: Tuple[int, int] | None,
    result_contrast: float,
    result_brightness: float,
    result_saturation: float,
) -> None:
    """It should return the successful result of an image capture with valid parameters."""
    subject = CaptureImageImpl(
        state_view=state_view,
        file_provider=file_provider,
        camera_provider=camera_provider_image_capture,
    )
    params = CaptureImageParams(
        fileName="coolpic",
        resolution=resolution,
        zoom=zoom,
        pan=pan,
        contrast=contrast,
        brightness=brightness,
        saturation=saturation,
    )
    decoy.when(state_view.files.get_filecount()).then_return(0)

    with mock.patch("os.path.exists", mock.Mock(return_value=True)):
        result = await subject.execute(params=params)
        assert result == SuccessData(
            public=CaptureImageResult(
                fileId=None,
                resolution=result_resolution,
                zoom=result_zoom,
                pan=result_pan,
                contrast=result_contrast,
                brightness=result_brightness,
                saturation=result_saturation,
            ),
            state_update=result.state_update,
        )


async def test_capture_image_result_has_clean_defaults(
    decoy: Decoy,
    state_view: StateView,
    file_provider: FileProvider,
    camera_provider_image_capture: CameraProvider,
) -> None:
    """It should return the successful result of an image capture with all expected defaults."""
    subject = CaptureImageImpl(
        state_view=state_view,
        file_provider=file_provider,
        camera_provider=camera_provider_image_capture,
    )
    params = CaptureImageParams(
        fileName="coolpic",
        resolution=None,
        zoom=None,
        pan=None,
        contrast=None,
        brightness=None,
        saturation=None,
    )
    decoy.when(state_view.files.get_filecount()).then_return(0)

    with mock.patch("os.path.exists", mock.Mock(return_value=True)):
        result = await subject.execute(params=params)
        assert result == SuccessData(
            public=CaptureImageResult(
                fileId=None,
                resolution=(1920, 1080),
                zoom=1.0,
                pan=None,
                contrast=50,
                brightness=50,
                saturation=50,
            ),
            state_update=result.state_update,
        )


async def test_raises_filename_error(
    decoy: Decoy,
    state_view: StateView,
    file_provider: FileProvider,
    camera_provider: CameraProvider,
) -> None:
    """It should raise FileNameInvalidError when the capture image command is provided a bad file name."""
    subject = CaptureImageImpl(
        state_view=state_view,
        file_provider=file_provider,
        camera_provider=camera_provider,
    )
    for char in SPECIAL_CHARACTERS:
        params = CaptureImageParams(fileName="badname" + char)
        with pytest.raises(FileNameInvalidError):
            await subject.execute(params=params)


@pytest.mark.parametrize(
    argnames=[
        "zoom",
        "contrast",
        "brightness",
        "saturation",
        "resolution",
    ],
    argvalues=[
        [0.9, 1, 1, 1, (1920, 1080)],
        [2.1, 1, 1, 1, (1920, 1080)],
        [1, -1, 1, 1, (1920, 1080)],
        [1, 101, 1, 1, (1920, 1080)],
        [1, 1, -1, 1, (1920, 1080)],
        [1, 1, 101, 1, (1920, 1080)],
        [1, 1, 1, -1, (1920, 1080)],
        [1, 1, 1, 101, (1920, 1080)],
        [1, 1, 1, 1, (0, 0)],
        [1, 1, 1, 1, (10000, 10000)],
    ],
)
async def test_raises_image_parameter_error(
    decoy: Decoy,
    state_view: StateView,
    file_provider: FileProvider,
    camera_provider_image_capture: CameraProvider,
    zoom: float,
    contrast: float,
    brightness: float,
    saturation: float,
    resolution: Tuple[int, int],
) -> None:
    """It should raise CameraSettingsInvalidError when the capture image command is provided bad filter params."""
    subject = CaptureImageImpl(
        state_view=state_view,
        file_provider=file_provider,
        camera_provider=camera_provider_image_capture,
    )
    params = CaptureImageParams(
        resolution=resolution,
        zoom=zoom,
        contrast=contrast,
        brightness=brightness,
        saturation=saturation,
    )

    decoy.when(state_view.files.get_filecount()).then_return(0)

    with mock.patch("os.path.exists", mock.Mock(return_value=True)):
        with pytest.raises(CameraSettingsInvalidError):
            await subject.execute(params=params)


async def test_raises_bad_resolution_and_zoom(
    state_view: StateView,
    file_provider: FileProvider,
    camera_provider: CameraProvider,
) -> None:
    """It should raise CameraSettingsInvalidError when the capture image command is provided a bad resolution or zoom, even when the camera callback is unavailable."""
    subject = CaptureImageImpl(
        state_view=state_view,
        file_provider=file_provider,
        camera_provider=camera_provider,
    )

    params = CaptureImageParams(resolution=(319, 239))
    with pytest.raises(CameraSettingsInvalidError):
        await subject.execute(params=params)
    params = CaptureImageParams(resolution=(7681, 4321))
    with pytest.raises(CameraSettingsInvalidError):
        await subject.execute(params=params)

    params = CaptureImageParams(zoom=0.9)
    with pytest.raises(CameraSettingsInvalidError):
        await subject.execute(params=params)
    params = CaptureImageParams(zoom=2.1)
    with pytest.raises(CameraSettingsInvalidError):
        await subject.execute(params=params)



"""Test Heater-Shaker start set temperature command implementation."""

from decoy import Decoy, matchers

from opentrons.hardware_control.modules import HeaterShaker
from opentrons.protocol_engine.actions import Action, ActionDispatcher, StartTaskAction
from opentrons.protocol_engine.commands import heater_shaker
from opentrons.protocol_engine.commands.command import SuccessData
from opentrons.protocol_engine.commands.heater_shaker.set_target_temperature import (
    SetTargetTemperatureImpl,
)
from opentrons.protocol_engine.execution import EquipmentHandler, TaskHandler
from opentrons.protocol_engine.resources import ModelUtils
from opentrons.protocol_engine.state.module_substates import (
    HeaterShakerModuleId,
    HeaterShakerModuleSubState,
)
from opentrons.protocol_engine.state.state import StateView
from opentrons.protocol_engine.types.tasks import Task


async def test_set_target_temperature(
    decoy: Decoy,
    state_view: StateView,
    equipment: EquipmentHandler,
    real_task_handler: TaskHandler,
    action_dispatcher: ActionDispatcher,
    model_utils: ModelUtils,
) -> None:
    """It should be able to set the specified module's target temperature."""
    subject = SetTargetTemperatureImpl(
        state_view=state_view, equipment=equipment, task_handler=real_task_handler
    )

    data = heater_shaker.SetTargetTemperatureParams(
        moduleId="input-heater-shaker-id", celsius=12.3, taskId="taskId"
    )

    hs_module_substate = decoy.mock(cls=HeaterShakerModuleSubState)
    hs_hardware = decoy.mock(cls=HeaterShaker)

    decoy.when(
        state_view.modules.get_heater_shaker_module_substate(
            module_id="input-heater-shaker-id"
        )
    ).then_return(hs_module_substate)

    decoy.when(model_utils.ensure_id("taskId")).then_return("taskId")
    decoy.when(hs_module_substate.module_id).then_return(
        HeaterShakerModuleId("heater-shaker-id")
    )

    # Stub temperature validation from hs module view
    decoy.when(
        hs_module_substate.validate_target_temperature(celsius=12.3)
    ).then_return(45.6)

    # Get attached hardware modules
    decoy.when(
        equipment.get_module_hardware_api(HeaterShakerModuleId("heater-shaker-id"))
    ).then_return(hs_hardware)

    task: Task | None = None

    def _capture_task(action: Action) -> None:
        nonlocal task
        assert isinstance(action, StartTaskAction)
        task = action.task

    decoy.when(
        action_dispatcher.dispatch(StartTaskAction(task=matchers.Anything()))  # type: ignore[func-returns-value]
    ).then_do(_capture_task)
    result = await subject.execute(data)
    assert task is not None
    await task.asyncioTask
    decoy.verify(await hs_hardware.start_set_temperature(celsius=45.6), times=1)
    decoy.verify(await hs_hardware.await_temperature(awaiting_temperature=45.6))
    assert result == SuccessData(
        public=heater_shaker.SetTargetTemperatureResult(taskId="taskId"),
    )



"""Tests for the engine based Protocol API module core implementations."""

from typing import cast

import pytest
from _pytest.fixtures import SubRequest
from decoy import Decoy

from ... import versions_at_or_above, versions_below
from opentrons.drivers.types import ThermocyclerLidStatus
from opentrons.hardware_control import SynchronousAdapter
from opentrons.hardware_control.modules import (
    ModuleType,
    TemperatureStatus,
    Thermocycler,
)
from opentrons.protocol_api import MAX_SUPPORTED_VERSION
from opentrons.protocol_api.core.engine.module_core import ThermocyclerModuleCore
from opentrons.protocol_api.core.engine.protocol import ProtocolCore
from opentrons.protocol_api.core.engine.tasks import EngineTaskCore
from opentrons.protocol_engine import commands as cmd
from opentrons.protocol_engine.clients import SyncClient as EngineClient
from opentrons.protocols.api_support.types import APIVersion

SyncThermocyclerHardware = SynchronousAdapter[Thermocycler]


@pytest.fixture
def mock_engine_client(decoy: Decoy) -> EngineClient:
    """Get a mock ProtocolEngine synchronous client."""
    return decoy.mock(cls=EngineClient)


@pytest.fixture
def mock_sync_module_hardware(decoy: Decoy) -> SyncThermocyclerHardware:
    """Get a mock synchronous module hardware."""
    return decoy.mock(name="SyncThermocyclerHardware")  # type: ignore[no-any-return]


@pytest.fixture
def mock_protocol_core(decoy: Decoy) -> ProtocolCore:
    """Get a mock protocol core."""
    return decoy.mock(cls=ProtocolCore)


@pytest.fixture
def subject(
    mock_engine_client: EngineClient,
    mock_sync_module_hardware: SyncThermocyclerHardware,
    mock_protocol_core: ProtocolCore,
) -> ThermocyclerModuleCore:
    """Get a ThermocyclerModuleCore test subject."""
    return ThermocyclerModuleCore(
        module_id="1234",
        engine_client=mock_engine_client,
        api_version=MAX_SUPPORTED_VERSION,
        sync_module_hardware=mock_sync_module_hardware,
        protocol_core=mock_protocol_core,
    )


@pytest.fixture(params=versions_below(APIVersion(2, 21), flex_only=False))
def subject_below_221(
    request: SubRequest,
    mock_engine_client: EngineClient,
    mock_sync_module_hardware: SyncThermocyclerHardware,
    mock_protocol_core: ProtocolCore,
) -> ThermocyclerModuleCore:
    """Get a ThermocyclerCore below API version 2.21."""
    return ThermocyclerModuleCore(
        module_id="1234",
        engine_client=mock_engine_client,
        api_version=cast(APIVersion, request.param),
        sync_module_hardware=mock_sync_module_hardware,
        protocol_core=mock_protocol_core,
    )


@pytest.fixture(params=versions_at_or_above(APIVersion(2, 21)))
def subject_at_or_above_221(
    request: SubRequest,
    mock_engine_client: EngineClient,
    mock_sync_module_hardware: SyncThermocyclerHardware,
    mock_protocol_core: ProtocolCore,
) -> ThermocyclerModuleCore:
    """Get a ThermocyclerCore below API version 2.21."""
    return ThermocyclerModuleCore(
        module_id="1234",
        engine_client=mock_engine_client,
        api_version=cast(APIVersion, request.param),
        sync_module_hardware=mock_sync_module_hardware,
        protocol_core=mock_protocol_core,
    )


def test_create(
    decoy: Decoy,
    mock_engine_client: EngineClient,
    mock_sync_module_hardware: SyncThermocyclerHardware,
    mock_protocol_core: ProtocolCore,
) -> None:
    """It should be able to create a thermocycler module core."""
    result = ThermocyclerModuleCore(
        module_id="1234",
        engine_client=mock_engine_client,
        api_version=MAX_SUPPORTED_VERSION,
        sync_module_hardware=mock_sync_module_hardware,
        protocol_core=mock_protocol_core,
    )

    assert result.module_id == "1234"
    assert result.MODULE_TYPE == ModuleType.THERMOCYCLER


def test_open_lid(
    decoy: Decoy, mock_engine_client: EngineClient, subject: ThermocyclerModuleCore
) -> None:
    """It should open the lid with the engine client."""
    result = subject.open_lid()

    decoy.verify(
        mock_engine_client.execute_command(
            cmd.thermocycler.OpenLidParams(moduleId="1234")
        ),
        times=1,
    )
    assert result == ThermocyclerLidStatus.OPEN


def test_close_lid(
    decoy: Decoy, mock_engine_client: EngineClient, subject: ThermocyclerModuleCore
) -> None:
    """It should close the lid with the engine client."""
    result = subject.close_lid()

    decoy.verify(
        mock_engine_client.execute_command(
            cmd.thermocycler.CloseLidParams(moduleId="1234")
        ),
        times=1,
    )
    assert result == ThermocyclerLidStatus.CLOSED


def test_set_target_block_temperature(
    decoy: Decoy, mock_engine_client: EngineClient, subject: ThermocyclerModuleCore
) -> None:
    """It should set the block temperature with the engine client."""
    subject.set_target_block_temperature(
        celsius=42.0,
        hold_time_seconds=1.2,
        block_max_volume=3.4,
        ramp_rate=None,
    )

    decoy.verify(
        mock_engine_client.execute_command(
            cmd.thermocycler.SetTargetBlockTemperatureParams(
                moduleId="1234",
                celsius=42.0,
                blockMaxVolumeUl=3.4,
                holdTimeSeconds=1.2,
                ramp_rate=None,
            )
        ),
        times=1,
    )


def test_start_set_target_block_temperature(
    decoy: Decoy, mock_engine_client: EngineClient, subject: ThermocyclerModuleCore
) -> None:
    """It should start to set the block temperature with the engine client and return an EngineTaskCore."""
    task_mock = decoy.mock(cls=EngineTaskCore)
    decoy.when(
        mock_engine_client.execute_command_without_recovery(
            cmd.thermocycler.SetTargetBlockTemperatureParams(
                moduleId="1234",
                celsius=42.0,
                blockMaxVolumeUl=3.4,
                ramp_rate=None,
            )
        )
    ).then_return(
        cmd.thermocycler.SetTargetBlockTemperatureResult(
            targetBlockTemperature=42.0, taskId="taskId"
        )
    )
    task_mock._id = "taskId"
    result = subject.start_set_target_block_temperature(
        celsius=42.0,
        block_max_volume=3.4,
        ramp_rate=None,
    )
    assert isinstance(result, EngineTaskCore)
    assert result._id == "taskId"


def test_wait_for_block_temperature(
    decoy: Decoy, mock_engine_client: EngineClient, subject: ThermocyclerModuleCore
) -> None:
    """It should wait for the block temperature with the engine client."""
    subject.wait_for_block_temperature()

    decoy.verify(
        mock_engine_client.execute_command(
            cmd.thermocycler.WaitForBlockTemperatureParams(moduleId="1234")
        ),
        times=1,
    )


def test_set_target_lid_temperature(
    decoy: Decoy,
    mock_engine_client: EngineClient,
    subject: ThermocyclerModuleCore,
) -> None:
    """It should set the lid temperature with the engine client."""
    subject.set_target_lid_temperature(celsius=42.0)

    decoy.verify(
        mock_engine_client.execute_command(
            cmd.thermocycler.SetTargetLidTemperatureParams(
                moduleId="1234", celsius=42.0
            )
        ),
        times=1,
    )


def test_start_set_target_lid_temperature(
    decoy: Decoy,
    mock_engine_client: EngineClient,
    subject: ThermocyclerModuleCore,
) -> None:
    """It should set the lid temperature with the engine client and return an EngineTaskCore."""
    task_mock = decoy.mock(cls=EngineTaskCore)
    decoy.when(
        mock_engine_client.execute_command_without_recovery(
            cmd.thermocycler.SetTargetLidTemperatureParams(
                moduleId="1234", celsius=42.0
            )
        ),
    ).then_return(
        cmd.thermocycler.SetTargetLidTemperatureResult(
            targetLidTemperature=42.0, taskId="taskId"
        )
    )
    task_mock._id = "taskId"
    result = subject.start_set_target_lid_temperature(42.0)
    assert isinstance(result, EngineTaskCore)
    assert result._id == "taskId"


def test_wait_for_lid_temperature(
    decoy: Decoy, mock_engine_client: EngineClient, subject: ThermocyclerModuleCore
) -> None:
    """It should wait for the lid temperature with the engine client."""
    subject.wait_for_lid_temperature()

    decoy.verify(
        mock_engine_client.execute_command(
            cmd.thermocycler.WaitForLidTemperatureParams(moduleId="1234")
        ),
        times=1,
    )


def test_execute_profile_below_221(
    decoy: Decoy,
    mock_engine_client: EngineClient,
    subject_below_221: ThermocyclerModuleCore,
) -> None:
    """It should run a thermocycler profile with the engine client."""
    subject_below_221.execute_profile(
        steps=[{"temperature": 45.6, "hold_time_seconds": 12.3, "ramp_rate": 0.0}],
        repetitions=2,
        block_max_volume=78.9,
    )

    decoy.verify(
        mock_engine_client.execute_command(
            cmd.thermocycler.RunProfileParams(
                moduleId="1234",
                profile=[
                    cmd.thermocycler.RunProfileStepParams(
                        celsius=45.6, holdSeconds=12.3, rampRate=0.0
                    ),
                    cmd.thermocycler.RunProfileStepParams(
                        celsius=45.6, holdSeconds=12.3, rampRate=0.0
                    ),
                ],
                blockMaxVolumeUl=78.9,
            )
        )
    )


def test_execute_profile_above_221(
    decoy: Decoy,
    mock_engine_client: EngineClient,
    subject_at_or_above_221: ThermocyclerModuleCore,
) -> None:
    """It should run a thermocycler profile with the engine client."""
    subject_at_or_above_221.execute_profile(
        steps=[
            {"temperature": 45.6, "hold_time_seconds": 12.3, "ramp_rate": 0.0},
            {"temperature": 78.9, "hold_time_seconds": 45.6, "ramp_rate": 1.0},
        ],
        repetitions=2,
        block_max_volume=25,
    )
    decoy.verify(
        mock_engine_client.execute_command(
            cmd.thermocycler.RunExtendedProfileParams(
                moduleId="1234",
                profileElements=[
                    cmd.thermocycler.ProfileCycle(
                        repetitions=2,
                        steps=[
                            cmd.thermocycler.ProfileStep(
                                celsius=45.6, holdSeconds=12.3, rampRate=0.0
                            ),
                            cmd.thermocycler.ProfileStep(
                                celsius=78.9, holdSeconds=45.6, rampRate=1.0
                            ),
                        ],
                    )
                ],
                blockMaxVolumeUl=25,
            )
        )
    )


def test_start_execute_profile(
    decoy: Decoy,
    mock_engine_client: EngineClient,
    subject: ThermocyclerModuleCore,
) -> None:
    """It should run a thermocycler profile with the engine client and return an EngineTaskCore."""
    task_mock = decoy.mock(cls=EngineTaskCore)
    decoy.when(
        mock_engine_client.execute_command_without_recovery(
            cmd.thermocycler.StartRunExtendedProfileParams(
                moduleId="1234",
                profileElements=[
                    cmd.thermocycler.ProfileCycle(
                        repetitions=2,
                        steps=[
                            cmd.thermocycler.ProfileStep(
                                celsius=45.6, holdSeconds=12.3, rampRate=0.0
                            ),
                            cmd.thermocycler.ProfileStep(
                                celsius=78.9, holdSeconds=45.6, rampRate=1.0
                            ),
                        ],
                    )
                ],
                blockMaxVolumeUl=25,
            )
        )
    ).then_return(cmd.thermocycler.StartRunExtendedProfileResult(taskId="taskId"))
    task_mock._id = "taskId"
    result = subject.start_execute_profile(
        steps=[
            {"temperature": 45.6, "hold_time_seconds": 12.3, "ramp_rate": 0.0},
            {"temperature": 78.9, "hold_time_seconds": 45.6, "ramp_rate": 1.0},
        ],
        repetitions=2,
        block_max_volume=25,
    )
    assert isinstance(result, EngineTaskCore)
    assert result._id == "taskId"


def test_deactivate_lid(
    decoy: Decoy, mock_engine_client: EngineClient, subject: ThermocyclerModuleCore
) -> None:
    """It should turn of the heated lid with the engine client."""
    subject.deactivate_lid()

    decoy.verify(
        mock_engine_client.execute_command(
            cmd.thermocycler.DeactivateLidParams(moduleId="1234")
        ),
        times=1,
    )


def test_deactivate_block(
    decoy: Decoy, mock_engine_client: EngineClient, subject: ThermocyclerModuleCore
) -> None:
    """It should turn of the well block temperature controller with the engine client."""
    subject.deactivate_block()

    decoy.verify(
        mock_engine_client.execute_command(
            cmd.thermocycler.DeactivateBlockParams(moduleId="1234")
        ),
        times=1,
    )


def test_deactivate(
    decoy: Decoy, mock_engine_client: EngineClient, subject: ThermocyclerModuleCore
) -> None:
    """It should turn of the well block temperature controller with the engine client."""
    subject.deactivate()

    decoy.verify(
        mock_engine_client.execute_command(
            cmd.thermocycler.DeactivateBlockParams(moduleId="1234")
        ),
        mock_engine_client.execute_command(
            cmd.thermocycler.DeactivateLidParams(moduleId="1234")
        ),
    )


def test_get_lid_position(
    decoy: Decoy,
    mock_sync_module_hardware: SyncThermocyclerHardware,
    subject: ThermocyclerModuleCore,
) -> None:
    """It should report the current lid position."""
    decoy.when(mock_sync_module_hardware.lid_status).then_return(
        ThermocyclerLidStatus.OPEN
    )
    result = subject.get_lid_position()
    assert result == ThermocyclerLidStatus.OPEN


def test_get_block_temperature_status(
    decoy: Decoy,
    mock_sync_module_hardware: SyncThermocyclerHardware,
    subject: ThermocyclerModuleCore,
) -> None:
    """It should report the current block temperature status."""
    decoy.when(mock_sync_module_hardware.status).then_return(TemperatureStatus.IDLE)
    result = subject.get_block_temperature_status()
    assert result == TemperatureStatus.IDLE


def test_get_lid_temperature_status(
    decoy: Decoy,
    mock_sync_module_hardware: SyncThermocyclerHardware,
    subject: ThermocyclerModuleCore,
) -> None:
    """It should report the current lid temperature status."""
    decoy.when(mock_sync_module_hardware.lid_temp_status).then_return(
        TemperatureStatus.IDLE
    )
    result = subject.get_lid_temperature_status()
    assert result == TemperatureStatus.IDLE


def test_get_block_temperature(
    decoy: Decoy,
    mock_sync_module_hardware: SyncThermocyclerHardware,
    subject: ThermocyclerModuleCore,
) -> None:
    """It should report the current block temperature."""
    decoy.when(mock_sync_module_hardware.temperature).then_return(12.3)
    result = subject.get_block_temperature()
    assert result == 12.3


def test_get_block_target_temperature(
    decoy: Decoy,
    mock_sync_module_hardware: SyncThermocyclerHardware,
    subject: ThermocyclerModuleCore,
) -> None:
    """It should report the target block temperature."""
    decoy.when(mock_sync_module_hardware.target).then_return(12.3)
    result = subject.get_block_target_temperature()
    assert result == 12.3


def test_get_lid_temperature(
    decoy: Decoy,
    mock_sync_module_hardware: SyncThermocyclerHardware,
    subject: ThermocyclerModuleCore,
) -> None:
    """It should report the current lid temperature."""
    decoy.when(mock_sync_module_hardware.lid_temp).then_return(42.0)
    result = subject.get_lid_temperature()
    assert result == 42.0


def test_get_lid_target_temperature(
    decoy: Decoy,
    mock_sync_module_hardware: SyncThermocyclerHardware,
    subject: ThermocyclerModuleCore,
) -> None:
    """It should report the current lid temperature."""
    decoy.when(mock_sync_module_hardware.lid_target).then_return(42.0)
    result = subject.get_lid_target_temperature()
    assert result == 42.0


def test_get_ramp_rate(
    decoy: Decoy,
    mock_sync_module_hardware: SyncThermocyclerHardware,
    subject: ThermocyclerModuleCore,
) -> None:
    """It should report the ramp rate."""
    decoy.when(mock_sync_module_hardware.ramp_rate).then_return(1.23)
    result = subject.get_ramp_rate()
    assert result == 1.23


def test_get_hold_time(
    decoy: Decoy,
    mock_sync_module_hardware: SyncThermocyclerHardware,
    subject: ThermocyclerModuleCore,
) -> None:
    """It should report the hold time."""
    decoy.when(mock_sync_module_hardware.hold_time).then_return(13.37)
    result = subject.get_hold_time()
    assert result == 13.37


def test_cycle_counting(
    decoy: Decoy,
    mock_sync_module_hardware: SyncThermocyclerHardware,
    subject: ThermocyclerModuleCore,
) -> None:
    """It should keep track of cycle and step counts and indices."""
    subject.execute_profile(
        [
            {"temperature": 45.6, "hold_time_seconds": 12.3, "ramp_rate": 0.0},
            {"temperature": 78.9, "hold_time_seconds": 45.6, "ramp_rate": 0.0},
        ],
        repetitions=3,
    )

    decoy.when(mock_sync_module_hardware.current_step_index).then_return(6)

    assert subject.get_total_cycle_count() == 3
    assert subject.get_current_cycle_index() == 3
    assert subject.get_total_step_count() == 2
    assert subject.get_current_step_index() == 2

    subject.deactivate_block()

    assert subject.get_total_cycle_count() is None
    assert subject.get_current_cycle_index() is None
    assert subject.get_total_step_count() is None
    assert subject.get_current_step_index() is None


def test_get_serial_number(
    decoy: Decoy, subject: ThermocyclerModuleCore, mock_engine_client: EngineClient
) -> None:
    """It should return a serial number."""
    decoy.when(mock_engine_client.state.modules.get_serial_number("1234")).then_return(
        "abc"
    )

    assert subject.get_serial_number() == "abc"



from opentrons import protocol_api

# metadata
metadata = {
    "protocolName": "Bug 7552",
    "author": "Name <email@address.com>",
    "description": "Simulation allows aspirating and dispensing on a tip rack",
    "apiLevel": "2.11",
}


# protocol run function. the part after the colon lets your editor know
# where to look for autocomplete suggestions
def run(protocol: protocol_api.ProtocolContext) -> None:
    # labware
    plate = protocol.load_labware("geb_96_tiprack_10ul", 4)

    # pipettes
    pipette = protocol.load_instrument("p300_single", "left", tip_racks=[plate])

    # commands
    pipette.transfer(5, plate.wells_by_name()["A1"], plate.wells_by_name()["B1"])



"""Test Protocol Engine motion planning utility methods."""

from typing import List, Optional

import pytest

from opentrons.motion_planning.types import MoveType
from opentrons.protocol_engine.state import _move_types as subject
from opentrons.protocol_engine.types import CurrentWell
from opentrons.types import Point


@pytest.mark.parametrize(
    ["location", "force_direct", "expected_move_type"],
    [
        (None, False, MoveType.GENERAL_ARC),
        (None, True, MoveType.DIRECT),
        (
            CurrentWell("other-pipette-id", "other-labware-id", "well-name"),
            False,
            MoveType.GENERAL_ARC,
        ),
        (
            CurrentWell("other-pipette-id", "labware-id", "well-name"),
            False,
            MoveType.GENERAL_ARC,
        ),
        (
            CurrentWell("pipette-id", "other-labware-id", "well-name"),
            False,
            MoveType.GENERAL_ARC,
        ),
        (CurrentWell("pipette-id", "labware-id", "well-name"), False, MoveType.DIRECT),
        (
            CurrentWell("pipette-id", "labware-id", "other-well-name"),
            False,
            MoveType.IN_LABWARE_ARC,
        ),
    ],
)
def test_get_move_type_to_well(
    location: Optional[CurrentWell], force_direct: bool, expected_move_type: MoveType
) -> None:
    """It should get the move type for a move to well operation."""
    move_type = subject.get_move_type_to_well(
        "pipette-id", "labware-id", "well-name", location, force_direct
    )

    assert move_type == expected_move_type


@pytest.mark.parametrize(
    ["edge_path_type", "expected_result"],
    [
        (
            subject.EdgePathType.LEFT,
            [
                Point(8, 20, 30),
                Point(10, 20, 30),
                Point(10, 27, 30),
                Point(10, 13, 30),
                Point(10, 20, 30),
            ],
        ),
        (
            subject.EdgePathType.RIGHT,
            [
                Point(12, 20, 30),
                Point(10, 20, 30),
                Point(10, 27, 30),
                Point(10, 13, 30),
                Point(10, 20, 30),
            ],
        ),
        (
            subject.EdgePathType.DEFAULT,
            [
                Point(12, 20, 30),
                Point(8, 20, 30),
                Point(10, 20, 30),
                Point(10, 27, 30),
                Point(10, 13, 30),
                Point(10, 20, 30),
            ],
        ),
    ],
)
def test_get_edge_point_list(
    edge_path_type: subject.EdgePathType,
    expected_result: List[Point],
) -> None:
    """It should get a list of well edge points."""
    result = subject.get_edge_point_list(
        Point(x=10, y=20, z=30),
        x_radius=5,
        y_radius=10,
        mm_from_edge=3,
        edge_path_type=edge_path_type,
    )

    assert result == expected_result



"""Unit tests for `stringify`."""

from decoy import Decoy

from opentrons_shared_data.labware.labware_definition import LabwareDefinition2

from opentrons.protocol_api.core.engine import stringify as subject
from opentrons.protocol_engine.clients.sync_client import SyncClient
from opentrons.protocol_engine.types import (
    OFF_DECK_LOCATION,
    DeckSlotLocation,
    ModuleDefinition,
    ModuleLocation,
    OnLabwareLocation,
)
from opentrons.types import DeckSlotName


def _make_dummy_labware_definition(
    decoy: Decoy, display_name: str
) -> LabwareDefinition2:
    mock = decoy.mock(cls=LabwareDefinition2)
    decoy.when(mock.metadata.displayName).then_return(display_name)
    return mock


def _make_dummy_module_definition(decoy: Decoy, display_name: str) -> ModuleDefinition:
    mock = decoy.mock(cls=ModuleDefinition)
    decoy.when(mock.displayName).then_return(display_name)
    return mock


def test_well_on_labware_without_user_display_name(decoy: Decoy) -> None:
    """Test stringifying a well on a labware that doesn't have a user-defined label."""
    mock_client = decoy.mock(cls=SyncClient)
    decoy.when(
        mock_client.state.labware.get_user_specified_display_name("labware-id")
    ).then_return(None)
    decoy.when(mock_client.state.labware.get_definition("labware-id")).then_return(
        _make_dummy_labware_definition(decoy, "definition-display-name")
    )
    decoy.when(mock_client.state.labware.get_location("labware-id")).then_return(
        OFF_DECK_LOCATION
    )

    result = subject.well(
        engine_client=mock_client, well_name="well-name", labware_id="labware-id"
    )
    assert result == "well-name of definition-display-name on [off-deck]"


def test_well_on_labware_with_user_display_name(decoy: Decoy) -> None:
    """Test stringifying a well on a labware that does have a user-defined label."""
    mock_client = decoy.mock(cls=SyncClient)
    decoy.when(
        mock_client.state.labware.get_user_specified_display_name("labware-id")
    ).then_return("user-display-name")
    decoy.when(mock_client.state.labware.get_definition("labware-id")).then_return(
        _make_dummy_labware_definition(decoy, "definition-display-name")
    )
    decoy.when(mock_client.state.labware.get_location("labware-id")).then_return(
        OFF_DECK_LOCATION
    )

    result = subject.well(
        engine_client=mock_client, well_name="well-name", labware_id="labware-id"
    )
    assert result == "well-name of user-display-name on [off-deck]"


def test_well_on_labware_with_complicated_location(decoy: Decoy) -> None:
    """Test stringifying a well on a labware with a deeply-nested location."""
    mock_client = decoy.mock(cls=SyncClient)

    decoy.when(
        mock_client.state.labware.get_user_specified_display_name("labware-id-1")
    ).then_return(None)
    decoy.when(mock_client.state.labware.get_definition("labware-id-1")).then_return(
        _make_dummy_labware_definition(decoy, "lw-1-display-name")
    )
    decoy.when(mock_client.state.labware.get_location("labware-id-1")).then_return(
        OnLabwareLocation(labwareId="labware-id-2")
    )

    decoy.when(
        mock_client.state.labware.get_user_specified_display_name("labware-id-2")
    ).then_return(None)
    decoy.when(mock_client.state.labware.get_definition("labware-id-2")).then_return(
        _make_dummy_labware_definition(decoy, "lw-2-display-name")
    )
    decoy.when(mock_client.state.labware.get_location("labware-id-2")).then_return(
        ModuleLocation(moduleId="module-id")
    )

    decoy.when(mock_client.state.modules.get_definition("module-id")).then_return(
        _make_dummy_module_definition(decoy, "module-display-name")
    )
    decoy.when(mock_client.state.modules.get_location("module-id")).then_return(
        DeckSlotLocation(slotName=DeckSlotName.SLOT_C2)
    )

    result = subject.well(
        engine_client=mock_client, well_name="well-name", labware_id="labware-id-1"
    )
    assert (
        result
        == "well-name of lw-1-display-name on lw-2-display-name on module-display-name on slot C2"
    )



"""Test load module command."""

from typing import cast

import pytest
from decoy import Decoy

from opentrons_shared_data.deck import load as load_deck
from opentrons_shared_data.deck.types import (
    DeckDefinitionV5,
    SlotDefV3,
)
from opentrons_shared_data.robot.types import RobotType

from opentrons.hardware_control.modules import ModuleType
from opentrons.hardware_control.modules.types import (
    FlexStackerModuleModel,
    HeaterShakerModuleModel,
    MagneticModuleModel,
    TemperatureModuleModel,
    ThermocyclerModuleModel,
)
from opentrons.hardware_control.modules.types import (
    ModuleModel as HardwareModuleModel,
)
from opentrons.protocol_engine import ModuleModel as EngineModuleModel
from opentrons.protocol_engine.commands.command import SuccessData
from opentrons.protocol_engine.commands.load_module import (
    LoadModuleImplementation,
    LoadModuleParams,
    LoadModuleResult,
)
from opentrons.protocol_engine.errors import LocationIsOccupiedError
from opentrons.protocol_engine.execution import EquipmentHandler, LoadedModuleData
from opentrons.protocol_engine.state.state import StateView
from opentrons.protocol_engine.state.update_types import LoadModuleUpdate, StateUpdate
from opentrons.protocol_engine.types import (
    AddressableAreaLocation,
    DeckSlotLocation,
    ModuleModel,
)
from opentrons.protocols.api_support.deck_type import (
    STANDARD_OT2_DECK,
    STANDARD_OT3_DECK,
)
from opentrons.types import DeckSlotName


@pytest.mark.parametrize(
    "module_model,module_def_fixture_name,load_slot_name,addressable_area_name",
    [
        (
            ModuleModel.TEMPERATURE_MODULE_V2,
            "tempdeck_v2_def",
            DeckSlotName.SLOT_D1,
            "temperatureModuleV2D1",
        ),
        (
            ModuleModel.MAGNETIC_BLOCK_V1,
            "mag_block_v1_def",
            DeckSlotName.SLOT_D1,
            "magneticBlockV1D1",
        ),
        (
            ModuleModel.THERMOCYCLER_MODULE_V2,
            "thermocycler_v2_def",
            # only B1 provides addressable area for thermocycler v2
            # so we use it here with decoy
            DeckSlotName.SLOT_B1,
            "thermocyclerModuleV2B1",
        ),
        (
            ModuleModel.HEATER_SHAKER_MODULE_V1,
            "heater_shaker_v1_def",
            DeckSlotName.SLOT_D3,
            "heaterShakerModuleV1D3",
        ),
        (
            ModuleModel.ABSORBANCE_READER_V1,
            "abs_reader_v1_def",
            DeckSlotName.SLOT_D3,
            "absorbanceReaderV1D3",
        ),
        (
            ModuleModel.FLEX_STACKER_MODULE_V1,
            "flex_stacker_v1_def",
            DeckSlotName.SLOT_D3,
            "flexStackerModuleV1D4",
        ),
    ],
)
async def test_load_module_implementation(
    request: pytest.FixtureRequest,
    decoy: Decoy,
    equipment: EquipmentHandler,
    state_view: StateView,
    ot3_standard_deck_def: DeckDefinitionV5,
    module_model: ModuleModel,
    module_def_fixture_name: str,
    load_slot_name: DeckSlotName,
    addressable_area_name: str,
) -> None:
    """A loadModule command should have an execution implementation."""
    module_definition = request.getfixturevalue(module_def_fixture_name)
    subject = LoadModuleImplementation(equipment=equipment, state_view=state_view)
    data = LoadModuleParams(
        model=module_model,
        location=DeckSlotLocation(slotName=load_slot_name),
        moduleId="some-id",
    )

    decoy.when(state_view.modules.get_deck_supports_module_fixtures()).then_return(True)

    decoy.when(state_view.labware.get_deck_definition()).then_return(
        ot3_standard_deck_def
    )
    decoy.when(
        state_view.addressable_areas.get_cutout_id_by_deck_slot_name(load_slot_name)
    ).then_return("cutout" + load_slot_name.value)

    decoy.when(
        state_view.geometry.ensure_location_not_occupied(
            DeckSlotLocation(slotName=load_slot_name), addressable_area_name
        )
    ).then_return(DeckSlotLocation(slotName=load_slot_name))

    addressable_area_provided_by_module = (
        f"{module_model}{load_slot_name}"
        if module_model != ModuleModel.FLEX_STACKER_MODULE_V1
        else f"{module_model}{load_slot_name.value[0]}4"
    )
    decoy.when(
        state_view.modules.ensure_and_convert_module_fixture_location(
            deck_slot=data.location.slotName,
            model=data.model,
        )
    ).then_return(addressable_area_provided_by_module)

    if module_model == ModuleModel.MAGNETIC_BLOCK_V1:
        decoy.when(
            await subject._equipment.load_magnetic_block(
                model=module_model,
                location=AddressableAreaLocation(
                    addressableAreaName=addressable_area_provided_by_module
                ),
                module_id="some-id",
            )
        ).then_return(
            LoadedModuleData(
                module_id="module-id",
                serial_number="mod-serial",
                definition=module_definition,
            )
        )
    else:
        decoy.when(
            await subject._equipment.load_module(
                model=module_model,
                location=AddressableAreaLocation(
                    addressableAreaName=addressable_area_provided_by_module
                ),
                module_id="some-id",
            )
        ).then_return(
            LoadedModuleData(
                module_id="module-id",
                serial_number="mod-serial",
                definition=module_definition,
            )
        )

    result = await subject.execute(data)
    decoy.verify(
        state_view.addressable_areas.raise_if_area_not_in_deck_configuration(
            addressable_area_provided_by_module
        )
    )

    expected_state_update = StateUpdate(
        loaded_module=LoadModuleUpdate(
            module_id="module-id",
            definition=module_definition,
            requested_model=module_model,
            serial_number="mod-serial",
            slot_name=load_slot_name,
        )
    )

    assert result == SuccessData(
        public=LoadModuleResult(
            moduleId="module-id",
            serialNumber="mod-serial",
            model=module_model,
        ),
        state_update=expected_state_update,
    )


async def test_load_module_raises_if_location_occupied(
    decoy: Decoy,
    equipment: EquipmentHandler,
    state_view: StateView,
) -> None:
    """A loadModule command should raise an error if the location is occupied."""
    subject = LoadModuleImplementation(equipment=equipment, state_view=state_view)

    data = LoadModuleParams(
        model=ModuleModel.TEMPERATURE_MODULE_V2,
        location=DeckSlotLocation(slotName=DeckSlotName.SLOT_D1),
        moduleId="some-id",
    )
    decoy.when(state_view.modules.get_deck_supports_module_fixtures()).then_return(True)
    addressable_area_provided_by_module = (
        f"{data.model}{data.location.slotName.value}"
        if data.model != ModuleModel.FLEX_STACKER_MODULE_V1
        else f"{data.model}{data.location.slotName.value[0]}4"
    )
    decoy.when(
        state_view.modules.ensure_and_convert_module_fixture_location(
            deck_slot=data.location.slotName,
            model=data.model,
        )
    ).then_return(addressable_area_provided_by_module)

    deck_def = load_deck(STANDARD_OT3_DECK, 5)

    decoy.when(state_view.labware.get_deck_definition()).then_return(deck_def)
    decoy.when(
        state_view.addressable_areas.get_cutout_id_by_deck_slot_name(
            DeckSlotName.SLOT_D1
        )
    ).then_return("cutout" + DeckSlotName.SLOT_D1.value)

    decoy.when(
        state_view.geometry.ensure_location_not_occupied(
            location=DeckSlotLocation(slotName=DeckSlotName.SLOT_D1),
            desired_addressable_area=addressable_area_provided_by_module,
        )
    ).then_raise(LocationIsOccupiedError("Get your own spot!"))

    with pytest.raises(LocationIsOccupiedError):
        await subject.execute(data)


@pytest.mark.parametrize(
    (
        "requested_model",
        "engine_model",
        "deck_def",
        "slot_name",
        "robot_type",
    ),
    [
        (
            TemperatureModuleModel.TEMPERATURE_V2,
            EngineModuleModel.TEMPERATURE_MODULE_V2,
            load_deck(STANDARD_OT3_DECK, 5),
            DeckSlotName.SLOT_D2,
            "OT-3 Standard",
        ),
        (
            ThermocyclerModuleModel.THERMOCYCLER_V1,
            EngineModuleModel.THERMOCYCLER_MODULE_V1,
            load_deck(STANDARD_OT2_DECK, 5),
            DeckSlotName.SLOT_1,
            "OT-2 Standard",
        ),
        (
            ThermocyclerModuleModel.THERMOCYCLER_V2,
            EngineModuleModel.THERMOCYCLER_MODULE_V2,
            load_deck(STANDARD_OT3_DECK, 5),
            DeckSlotName.SLOT_A2,
            "OT-3 Standard",
        ),
        (
            HeaterShakerModuleModel.HEATER_SHAKER_V1,
            EngineModuleModel.HEATER_SHAKER_MODULE_V1,
            load_deck(STANDARD_OT3_DECK, 5),
            DeckSlotName.SLOT_A2,
            "OT-3 Standard",
        ),
        (
            FlexStackerModuleModel.FLEX_STACKER_V1,
            EngineModuleModel.FLEX_STACKER_MODULE_V1,
            load_deck(STANDARD_OT3_DECK, 5),
            DeckSlotName.SLOT_A2,
            "OT-3 Standard",
        ),
    ],
)
async def test_load_module_raises_wrong_location(
    decoy: Decoy,
    equipment: EquipmentHandler,
    state_view: StateView,
    requested_model: HardwareModuleModel,
    engine_model: EngineModuleModel,
    deck_def: DeckDefinitionV5,
    slot_name: DeckSlotName,
    robot_type: RobotType,
) -> None:
    """It should raise an error when trying to load a module in an incompatible location."""
    subject = LoadModuleImplementation(equipment=equipment, state_view=state_view)

    data = LoadModuleParams(
        model=engine_model,
        location=DeckSlotLocation(slotName=slot_name),
        moduleId="some-id",
    )

    decoy.when(state_view.config.robot_type).then_return(robot_type)

    if robot_type == "OT-2 Standard":
        decoy.when(
            state_view.addressable_areas.get_slot_definition(slot_name.id)
        ).then_return(cast(SlotDefV3, {"compatibleModuleTypes": []}))
    else:
        decoy.when(state_view.labware.get_deck_definition()).then_return(deck_def)
        decoy.when(
            state_view.addressable_areas.get_cutout_id_by_deck_slot_name(slot_name)
        ).then_return("cutout" + slot_name.value)

    with pytest.raises(
        ValueError,
        match=f"A {ModuleType.from_model(model=requested_model).value} cannot be loaded into slot {slot_name}",
    ):
        await subject.execute(data)


@pytest.mark.parametrize(
    (
        "requested_model",
        "engine_model",
        "deck_def",
        "slot_name",
        "robot_type",
    ),
    [
        (
            MagneticModuleModel.MAGNETIC_V2,
            EngineModuleModel.MAGNETIC_MODULE_V2,
            load_deck(STANDARD_OT3_DECK, 5),
            DeckSlotName.SLOT_A2,
            "OT-3 Standard",
        ),
    ],
)
async def test_load_module_raises_module_fixture_id_does_not_exist(
    decoy: Decoy,
    equipment: EquipmentHandler,
    state_view: StateView,
    requested_model: HardwareModuleModel,
    engine_model: EngineModuleModel,
    deck_def: DeckDefinitionV5,
    slot_name: DeckSlotName,
    robot_type: RobotType,
) -> None:
    """It should raise an error when the module fixture ID does not exist."""
    subject = LoadModuleImplementation(equipment=equipment, state_view=state_view)

    data = LoadModuleParams(
        model=engine_model,
        location=DeckSlotLocation(slotName=slot_name),
        moduleId="some-id",
    )

    decoy.when(state_view.config.robot_type).then_return(robot_type)

    if robot_type == "OT-2 Standard":
        decoy.when(
            state_view.addressable_areas.get_slot_definition(slot_name.id)
        ).then_return(cast(SlotDefV3, {"compatibleModuleTypes": []}))
    else:
        decoy.when(state_view.labware.get_deck_definition()).then_return(deck_def)
        decoy.when(
            state_view.addressable_areas.get_cutout_id_by_deck_slot_name(slot_name)
        ).then_return("cutout" + slot_name.value)

    with pytest.raises(
        ValueError,
        match=f"Module Type {ModuleType.from_model(requested_model).value} does not have a related fixture ID.",
    ):
        await subject.execute(data)



"""Tests for the Parameter Definitions."""

import inspect

import pytest
from decoy import Decoy

from opentrons.protocol_engine.types import (
    BooleanParameter,
    EnumChoice,
    EnumParameter,
    NumberParameter,
)
from opentrons.protocols.parameters import validation as mock_validation
from opentrons.protocols.parameters.exceptions import ParameterValueError
from opentrons.protocols.parameters.parameter_definition import (
    create_bool_parameter,
    create_float_parameter,
    create_int_parameter,
    create_str_parameter,
)


@pytest.fixture(autouse=True)
def _patch_parameter_validation(decoy: Decoy, monkeypatch: pytest.MonkeyPatch) -> None:
    for name, func in inspect.getmembers(mock_validation, inspect.isfunction):
        monkeypatch.setattr(mock_validation, name, decoy.mock(func=func))


def test_create_int_parameter_min_and_max(decoy: Decoy) -> None:
    """It should create an int parameter definition with a minimum and maximum."""
    decoy.when(mock_validation.ensure_display_name("foo")).then_return("my cool name")
    decoy.when(mock_validation.ensure_variable_name("bar")).then_return("my variable")
    decoy.when(mock_validation.ensure_description("a b c")).then_return("1 2 3")
    decoy.when(mock_validation.ensure_unit_string_length("test")).then_return("microns")

    parameter_def = create_int_parameter(
        display_name="foo",
        variable_name="bar",
        default=42,
        minimum=1,
        maximum=100,
        description="a b c",
        unit="test",
    )

    decoy.verify(
        mock_validation.validate_options(42, 1, 100, None, int),
        mock_validation.validate_type(42, int),
    )

    assert parameter_def._display_name == "my cool name"
    assert parameter_def.variable_name == "my variable"
    assert parameter_def._description == "1 2 3"
    assert parameter_def._unit == "microns"
    assert parameter_def._allowed_values is None
    assert parameter_def._minimum == 1
    assert parameter_def._maximum == 100
    assert parameter_def.value == 42


def test_create_int_parameter_choices(decoy: Decoy) -> None:
    """It should create an int parameter definition with choices."""
    decoy.when(mock_validation.ensure_display_name("foo")).then_return("my cool name")
    decoy.when(mock_validation.ensure_variable_name("bar")).then_return("my variable")
    decoy.when(mock_validation.ensure_description(None)).then_return("1 2 3")
    decoy.when(mock_validation.ensure_unit_string_length(None)).then_return("microns")

    parameter_def = create_int_parameter(
        display_name="foo",
        variable_name="bar",
        default=42,
        choices=[{"display_name": "uhh", "value": 42}],
    )

    decoy.verify(
        mock_validation.validate_options(
            42, None, None, [{"display_name": "uhh", "value": 42}], int
        ),
        mock_validation.validate_type(42, int),
    )

    assert parameter_def._display_name == "my cool name"
    assert parameter_def.variable_name == "my variable"
    assert parameter_def._description == "1 2 3"
    assert parameter_def._unit == "microns"
    assert parameter_def._allowed_values == {42}
    assert parameter_def._minimum is None
    assert parameter_def._maximum is None
    assert parameter_def.value == 42


def test_int_parameter_default_raises_not_in_range() -> None:
    """It should raise an error if the default is not between min or max"""
    with pytest.raises(ParameterValueError, match="between"):
        create_int_parameter(
            display_name="foo",
            variable_name="bar",
            default=9000,
            minimum=9001,
            maximum=10000,
        )


def test_create_float_parameter_min_and_max(decoy: Decoy) -> None:
    """It should create a float parameter definition with a minimum and maximum."""
    decoy.when(mock_validation.ensure_display_name("foo")).then_return("my cool name")
    decoy.when(mock_validation.ensure_variable_name("bar")).then_return("my variable")
    decoy.when(mock_validation.ensure_description("a b c")).then_return("1 2 3")
    decoy.when(mock_validation.ensure_unit_string_length("test")).then_return("microns")

    parameter_def = create_float_parameter(
        display_name="foo",
        variable_name="bar",
        default=4.2,
        minimum=1.0,
        maximum=10.5,
        description="a b c",
        unit="test",
    )

    decoy.verify(
        mock_validation.validate_options(4.2, 1.0, 10.5, None, float),
        mock_validation.validate_type(4.2, float),
    )

    assert parameter_def._display_name == "my cool name"
    assert parameter_def.variable_name == "my variable"
    assert parameter_def._description == "1 2 3"
    assert parameter_def._unit == "microns"
    assert parameter_def._allowed_values is None
    assert parameter_def._minimum == 1.0
    assert parameter_def._maximum == 10.5
    assert parameter_def.value == 4.2


def test_create_float_parameter_choices(decoy: Decoy) -> None:
    """It should create a float parameter definition with choices."""
    decoy.when(mock_validation.ensure_display_name("foo")).then_return("my cool name")
    decoy.when(mock_validation.ensure_variable_name("bar")).then_return("my variable")

    parameter_def = create_float_parameter(
        display_name="foo",
        variable_name="bar",
        default=4.2,
        choices=[{"display_name": "urr", "value": 4.2}],
    )

    decoy.verify(
        mock_validation.validate_options(
            4.2, None, None, [{"display_name": "urr", "value": 4.2}], float
        ),
        mock_validation.validate_type(4.2, float),
    )

    assert parameter_def._display_name == "my cool name"
    assert parameter_def.variable_name == "my variable"
    assert parameter_def._allowed_values == {4.2}
    assert parameter_def._minimum is None
    assert parameter_def._maximum is None
    assert parameter_def.value == 4.2


def test_float_parameter_default_raises_not_in_range() -> None:
    """It should raise an error if the default is not between min or max"""
    with pytest.raises(ParameterValueError, match="between"):
        create_float_parameter(
            display_name="foo",
            variable_name="bar",
            default=9000.1,
            minimum=1,
            maximum=9000,
        )


def test_create_bool_parameter(decoy: Decoy) -> None:
    """It should create a boolean parameter"""
    decoy.when(mock_validation.ensure_display_name("foo")).then_return("my cool name")
    decoy.when(mock_validation.ensure_variable_name("bar")).then_return("my variable")
    decoy.when(mock_validation.ensure_description("describe this")).then_return("1 2 3")

    parameter_def = create_bool_parameter(
        display_name="foo",
        variable_name="bar",
        default=False,
        choices=[{"display_name": "uhh", "value": False}],
        description="describe this",
    )

    decoy.verify(
        mock_validation.validate_options(
            False, None, None, [{"display_name": "uhh", "value": False}], bool
        ),
        mock_validation.validate_type(False, bool),
    )

    assert parameter_def._display_name == "my cool name"
    assert parameter_def.variable_name == "my variable"
    assert parameter_def._description == "1 2 3"
    assert parameter_def._unit is None
    assert parameter_def._allowed_values == {False}
    assert parameter_def._minimum is None
    assert parameter_def._maximum is None
    assert parameter_def.value is False


def test_create_str_parameter(decoy: Decoy) -> None:
    """It should create a string parameter"""
    decoy.when(mock_validation.ensure_display_name("foo")).then_return("my cool name")
    decoy.when(mock_validation.ensure_variable_name("bar")).then_return("my variable")
    decoy.when(mock_validation.ensure_description("describe this")).then_return("1 2 3")

    parameter_def = create_str_parameter(
        display_name="foo",
        variable_name="bar",
        default="omega",
        choices=[{"display_name": "alpha", "value": "omega"}],
        description="describe this",
    )

    decoy.verify(
        mock_validation.validate_options(
            "omega", None, None, [{"display_name": "alpha", "value": "omega"}], str
        ),
        mock_validation.validate_type("omega", str),
    )

    assert parameter_def._display_name == "my cool name"
    assert parameter_def.variable_name == "my variable"
    assert parameter_def._description == "1 2 3"
    assert parameter_def._unit is None
    assert parameter_def._allowed_values == {"omega"}
    assert parameter_def._minimum is None
    assert parameter_def._maximum is None
    assert parameter_def.value == "omega"


def test_str_parameter_default_raises_not_in_allowed_values() -> None:
    """It should raise an error if the default is not between min or max"""
    with pytest.raises(ParameterValueError, match="allowed values"):
        create_str_parameter(
            display_name="foo",
            variable_name="bar",
            default="waldo",
            choices=[{"display_name": "where's", "value": "odlaw"}],
        )


def test_as_protocol_engine_boolean_parameter(decoy: Decoy) -> None:
    """It should return a protocol engine BooleanParameter model."""
    decoy.when(mock_validation.ensure_display_name("foo")).then_return("my cool name")
    decoy.when(mock_validation.ensure_variable_name("bar")).then_return("my variable")
    decoy.when(mock_validation.ensure_description("describe this")).then_return("1 2 3")

    parameter_def = create_bool_parameter(
        display_name="foo",
        variable_name="bar",
        default=False,
        choices=[{"display_name": "uhh", "value": False}],
        description="describe this",
    )

    assert parameter_def.as_protocol_engine_type() == BooleanParameter(
        type="bool",
        displayName="my cool name",
        variableName="my variable",
        description="1 2 3",
        value=False,
        default=False,
    )


def test_as_protocol_engine_enum_parameter(decoy: Decoy) -> None:
    """It should return a protocol engine EnumParameter model."""
    decoy.when(mock_validation.ensure_display_name("foo")).then_return("my cool name")
    decoy.when(mock_validation.ensure_variable_name("bar")).then_return("my variable")

    parameter_def = create_str_parameter(
        display_name="foo",
        variable_name="bar",
        default="red",
        choices=[
            {"display_name": "Lapis lazuli", "value": "blue"},
            {"display_name": "Vermilion", "value": "red"},
            {"display_name": "Celadon", "value": "green"},
        ],
    )
    parameter_def.value = "green"
    decoy.when(mock_validation.convert_type_string_for_enum(str)).then_return("float")

    assert parameter_def.as_protocol_engine_type() == EnumParameter(
        type="float",
        displayName="my cool name",
        variableName="my variable",
        choices=[
            EnumChoice(displayName="Lapis lazuli", value="blue"),
            EnumChoice(displayName="Vermilion", value="red"),
            EnumChoice(displayName="Celadon", value="green"),
        ],
        value="green",
        default="red",
    )


def test_as_protocol_engine_number_parameter(decoy: Decoy) -> None:
    """It should return a protocol engine NumberParameter model."""
    decoy.when(mock_validation.ensure_display_name("foo")).then_return("my cool name")
    decoy.when(mock_validation.ensure_variable_name("bar")).then_return("my variable")
    decoy.when(mock_validation.ensure_description("a b c")).then_return("1 2 3")
    decoy.when(mock_validation.ensure_unit_string_length("test")).then_return("microns")

    parameter_def = create_int_parameter(
        display_name="foo",
        variable_name="bar",
        default=42,
        minimum=1,
        maximum=100,
        description="a b c",
        unit="test",
    )

    parameter_def.value = 60
    decoy.when(mock_validation.convert_type_string_for_num_param(int)).then_return(
        "int"
    )

    assert parameter_def.as_protocol_engine_type() == NumberParameter(
        type="int",
        displayName="my cool name",
        variableName="my variable",
        description="1 2 3",
        suffix="microns",
        min=1.0,
        max=100.0,
        value=60.0,
        default=42.0,
    )



from pathlib import Path

import numpy as np

from opentrons import calibration_storage, config
from opentrons.hardware_control import robot_calibration
from opentrons.hardware_control.instruments.ot2 import instrument_calibration
from opentrons.types import Mount, Point
from opentrons.util.helpers import utc_now


def test_migrate_affine_xy_to_attitude() -> None:
    affine = [
        [1.0, 2.0, 3.0, 4.0],
        [5.0, 6.0, 7.0, 8.0],
        [9.0, 10.0, 11.0, 12.0],
        [13.0, 14.0, 15.0, 16.0],
    ]

    expected = [[1.0, 2.0, 3.0], [5.0, 6.0, 7.0], [0.0, 0.0, 1.0]]

    result = robot_calibration.migrate_affine_xy_to_attitude(affine)
    assert result == expected


def test_save_calibration() -> None:
    pathway = (
        config.get_opentrons_path("robot_calibration_dir") / "deck_calibration.json"
    )
    pip_id = "fakePip"
    lw_hash = "fakeHash"
    e = ((1, 1, 3), (2, 2, 2), (1, 2, 1))
    a = ((1.1, 3.1, 1.1), (2.1, 2.1, 2.2), (1.1, 2.1, 1.1))
    transform = [[0.975, 0.05, 0.0], [-1.025, 1.05, 0.0], [0.0, 0.0, 1.0]]
    expected = {
        "attitude": transform,
        "pipette_calibrated_with": pip_id,
        "last_modified": None,
        "tiprack": lw_hash,
        "source": "user",
        "status": {"markedBad": False, "markedAt": None, "source": None},
    }
    robot_calibration.save_attitude_matrix(e, a, pip_id, lw_hash)
    data = calibration_storage.file_operators.read_cal_file(pathway)
    data["last_modified"] = None
    assert data == expected


def test_load_calibration() -> None:
    pathway = Path(config.get_opentrons_path("robot_calibration_dir"))

    data = {
        "attitude": [[1, 0, 1], [0, 1, -0.5], [0, 0, 1]],
        "pipette_calibrated_with": "fake",
        "last_modified": utc_now(),
        "tiprack": "hash",
    }
    calibration_storage.file_operators.save_to_file(pathway, "deck_calibration", data)
    obj = robot_calibration.load_attitude_matrix()
    transform = [[1.0, 0, 1], [0, 1, -0.5], [0.0, 0, 1]]
    assert np.allclose(obj.attitude, transform)


def test_load_malformed_calibration() -> None:
    pathway = Path(config.get_opentrons_path("robot_calibration_dir"))
    data = {
        "atsadasitude": [[1, 0, 1], [0, 1, -0.5], [0, 0, 1]],
        "last_modified": utc_now(),
        "tiprack": "hash",
        "statu": [1, 2, 3],
    }
    calibration_storage.file_operators.save_to_file(pathway, "deck_calibration", data)
    obj = robot_calibration.load_attitude_matrix()
    assert np.allclose(obj.attitude, [[1, 0, 0], [0, 1, 0], [0, 0, 1]])


def test_load_json() -> None:
    path = config.get_opentrons_path("robot_calibration_dir") / "deck_calibration.json"
    path.write_text("{")
    obj = robot_calibration.load_attitude_matrix()
    assert obj.attitude == [[1, 0, 0], [0, 1, 0], [0, 0, 1]]


def test_load_pipette_offset() -> None:
    pip_id = "fakePip"
    mount = Mount.LEFT
    offset = Point(1, 2, 3)

    calibration_storage.ot2.save_pipette_calibration(
        offset, pip_id, mount, "hash", "opentrons/opentrons_96_tiprack_10ul/1"
    )
    obj = instrument_calibration.load_pipette_offset(pip_id, mount)

    assert np.allclose(obj.offset, offset)


def test_load_bad_pipette_offset() -> None:
    path = config.get_opentrons_path("pipette_calibration_dir") / "left"
    path.mkdir(parents=True, exist_ok=True)
    calpath = path / "fakePip.json"
    calpath.write_text("{")
    obj = instrument_calibration.load_pipette_offset("fakePip", Mount.LEFT)
    assert obj.offset == Point(0, 0, 0)
